# -.- coding: utf-8 -.-

# Teamgeist
# bases.py - Base classes defining the interfaces of the Telepathy framework
#
# Copyright (C) 2005-2009 Collabora Limited
# *  Contact: Youness Alaoui
# Copyright (C) 2005 Nokia Corporation
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# * Contributors:
# *   Youness Alaoui, Collabora Ltd.
# *   David Laban, Collabora Ltd.

"""
This module contains the base classes for dbus and telepathy client objects.
If you are developing a new client, it might be useful to set the environment
variable TPPY_DEBUG=signals. This will cause unhandled dbus signals to be
printed.
"""
import os

import gobject
import dbus
import telepathy
import logging

from utils import do_reply

logger = logging.getLogger('telepathy:bases')

def default_error_handler(exception):
    logging.basicConfig()
    logger.warning('Exception from asynchronous method call:\n%s' % exception)

class AsyncCallError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class AsyncMethodBuilder(dbus.Interface):
    def __init__(self, obj, interface, method_names):
        dbus.Interface.__init__(self, obj, interface)
        self._method_names = method_names

    def __getattr__(self, name):
        if name in self._method_names:
            def AsyncMethod(*args, **kwargs):
                reply_handler = kwargs.pop('reply_handler', None)
                error_handler = kwargs.pop('error_handler', None)
                ignore_reply = kwargs.get('ignore_reply', False)
                ignore_ready = kwargs.pop('ignore_ready', False)
                force_sync = kwargs.pop('force_sync', False)

                if reply_handler:
                    kwargs['reply_handler'] = lambda *args: \
                        do_reply(reply_handler, *args)

                if error_handler:
                    kwargs['error_handler'] = lambda *args: \
                        do_reply(error_handler, *args)

                if force_sync is False and \
                        reply_handler is None and \
                        error_handler is None and \
                        ignore_reply is False:
                    raise AsyncCallError("Calling '%s' synchronously" % name)

                method = dbus.Interface.__getattr__(self, name)
                if force_sync is False and \
                        isinstance(self, InterfaceBuilder) and \
                        self.proxy_object.is_base_ready() is False and \
                        ignore_ready is False:
                    logger.info("Class base not ready yet, "
                                "defering method call '%s'"
                                % name)
                    def DeferedMethod():
                        logger.info("Deferred method '%s' called" % name)
                        return method(*args, **kwargs)
                    self.proxy_object.call_when_base_ready(DeferedMethod)
                else:
                    return method(*args, **kwargs)
            return AsyncMethod
        else:
            raise AttributeError, name


class InterfacePropertyAccessor(AsyncMethodBuilder):

    def get_property(self, name, **kwargs):
        """Use the Properties interface to get a property from an interface."""
        return self.proxy_object[dbus.PROPERTIES_IFACE].Get(
            self.dbus_interface, name, **kwargs)

    def get_all_properties(self, **kwargs):
        return self.proxy_object[dbus.PROPERTIES_IFACE].GetAll(
            self.dbus_interface, **kwargs)

    def set_property(self, name, value, **kwargs):
        """Use the Properties interface to set a property to an interface."""
        return self.proxy_object[dbus.PROPERTIES_IFACE].Set(
            self.dbus_interface, name, value, **kwargs)

class InterfaceBuilder(dbus.proxies.ProxyObject):
    """
    A base class which inherits `dbus.proxies.ProxyObject` and allows
    access to individual interfaces of the object.

    This object acts like a dict where the keys are D-Bus interface names,
    and the values are `dbus.Interface` objects.
    """

    # Method names that are available, must be overriden by derived classes
    _method_names = {}
    _dbus_properties_methods = ['Get', 'GetAll', 'Set']

    def __init__(self, main_interface, bus_name = None, object_path = None,
                 bus = None):

        if bus is None:
            bus = dbus.Bus()

        if bus_name is not None and object_path is None:
            object_path = '/' + bus_name.replace('.', '/')

        if bus_name is None and object_path is not None:
            bus_name = object_path.lstrip('/').replace('/', '.')

        if bus_name is None or object_path is None:
            raise TypeError("Bus name or object path should not be None")

        dbus.proxies.ProxyObject.__init__(self, bus, bus_name, object_path)

        self._main_interface = main_interface
        self._got_channel_type = False
        self._got_interfaces = False
        self._interfaces = set()
        self._interfaces.add(dbus.PROPERTIES_IFACE)
        self._interfaces.add(self._main_interface)
        self.__interface_objs = {}
        self.__ready = False

    def __getitem__(self, name):
        if name not in self.__interface_objs:
            if self.is_base_ready() and name not in self._interfaces:
                raise KeyError(name)
            if name == dbus.PROPERTIES_IFACE:
                method_names = self._dbus_properties_methods
            elif name in self._method_names:
                method_names = self._method_names[name]
            self.__interface_objs[name] = InterfacePropertyAccessor(self,
                                                                    name,
                                                                    method_names)

        return self.__interface_objs[name]

    def __getattr__(self, name):
        return getattr(self[self._main_interface], name)

    def __contains__(self, name):
        return name in self.__interface_objs or \
            (self.__ready and name in self._interfaces)

    def _get_interfaces_reply_cb(self, interfaces):
        if self.__ready:
            return

        self._interfaces.update(interfaces)
        self._got_interfaces = True

        if self._main_interface != telepathy.interfaces.CHANNEL or \
                self._got_channel_type:
            self.__ready = True

            if self.__ready_handler:
                self.__ready_handler()


    def _get_interfaces_error_cb(self, interfaces):
        try:
            self.GetInterfaces(
                reply_handler=self._get_interfaces_reply_cb,
                error_handler=lambda err: self._get_interfaces_reply_cb([]),
                ignore_ready=True)
        except AttributeError:
            self._get_interfaces_reply_cb([])

    def _get_channel_type_reply_cb(self, channel_type):
        if self.__ready:
            return

        if channel_type != None:
            self._interfaces.update([channel_type,])

        self._got_channel_type = True
        if self._got_interfaces:
            self.__ready = True

            if self.__ready_handler:
                self.__ready_handler()

    def _get_channel_type_error_cb(self, interfaces):
        try:
            self.GetChannelType(
                reply_handler=self._get_channel_type_reply_cb,
                error_handler=lambda err: self._get_channel_type_reply_cb(None),
                ignore_ready=True)
        except AttributeError:
            self._get_interfaces_reply_cb([])

    def base_ready(self, ready_handler):
        self.__ready_handler = ready_handler
        self.get_property("Interfaces",
                          reply_handler=self._get_interfaces_reply_cb,
                          error_handler=self._get_interfaces_error_cb,
                          ignore_ready=True)
        if self._main_interface == telepathy.interfaces.CHANNEL:
            self.get_property("ChannelType",
                              reply_handler=self._get_channel_type_reply_cb,
                              error_handler=self._get_channel_type_error_cb,
                              ignore_ready=True)

    def is_base_ready(self):
        return self.__ready

    def get_property(self, name, **kwargs):
        return self[self._main_interface].get_property(name, **kwargs)

    def get_all_properties(self, callback):
        return self[self._main_interface].get_all_properties(callback)

    def set_property(self, name, value, **kwargs):
        return self[self._main_interface].set_property(name, value, **kwargs)


class ObjectListener(InterfaceBuilder):
    """
    A base class for anything that listens to remote objects
    This implements __eq__ because dbus.Object doesn't do so.
    """

    # Signal names to listen to, must be overriden by derived classes
    _signal_names = {}

    def __init__(self, main_interface, bus_name,object_path = None, bus = None):
        InterfaceBuilder.__init__(self, main_interface, bus_name,
                                  object_path, bus)
        self.__signal_matches = []
        self.__override_getattr = False

    def __getattr__(self, name):
        if self.__override_getattr:
            raise AttributeError, name
        else:
            return super(ObjectListener, self).__getattr__(name)

    def _register_callbacks(self):
        """
        Registers callbacks for all signals provided by the connection.
        Inheret from this class and override the appropriate method to handle
        each event.
        """

        self.__override_getattr = True
        for iface_name in self._interfaces:
            if iface_name in self._signal_names:
                for signal_name in self._signal_names[iface_name]:
                    iface_suffix = iface_name.split('.')[-1]
                    if (signal_name.startswith('__') and
                        signal_name.endswith('__')):
                        callbacks = [signal_name]
                    else:
                        callbacks = ["%s_%s" % (iface_suffix, signal_name),
                                   signal_name]
                    matched = False
                    for cb_name in callbacks:
                        cb = getattr(self, cb_name, None)
                        if cb is not None:
                            signal_match = self[iface_name]. \
                                connect_to_signal(signal_name, cb)
                            self.__signal_matches.append(signal_match)
                            matched = True
                    if matched is False:
                        # FIXME: actually decide how to configure debug messages.
                        if "signals" in os.environ.get("TPPY_DEBUG", ""):
                            cb = self._make_printer(iface_name, signal_name)
                            self[iface_name].connect_to_signal(signal_name, cb)
        self.__override_getattr = False

    def disconnect_signal(self, signal):
        signal.remove()
        self.__signal_matches.remove(signal)

    def disconnect_all_signals(self):
        for signal in self.__signal_matches:
            self.disconnect_signal(signal)

    def _make_printer(self, iface_name, function_name):
        """
        Creates a callback function for debugging missing signal handlers.
        This is only called if there is no method of the appropriate name
        which can be used as a callback.
        """
        def printer(*args, **kwargs):
            logger.warning("Signal emitted: %s.%s(%s)" % (iface_name, function_name,
                    ', '.join(map(str, args))))
        return printer

class ReadyableObject(object):
    def __init__(self):
        self.__ready = False
        self._ready_handlers = []

    def call_when_ready(self, handler):
        if self.is_ready():
            do_reply(handler)
        else:
            self._ready_handlers.append(handler)

    def is_ready(self):
        return self.__ready

    def ready(self):
        if self.__ready is True:
            return

        self.__ready = True
        while(len(self._ready_handlers) > 0):
            handler = self._ready_handlers.pop(0)
            do_reply(handler)


class BaseTelepathyObject(ObjectListener):

    __object_counter = 0

    def __init__(self, main_interface, bus_name,object_path = None, bus = None):
        ObjectListener.__init__(self, main_interface, bus_name,
                                  object_path, bus)
        self.__base_ready = False
        self._base_ready_handlers = []
        BaseTelepathyObject.__object_counter += 1
        self.__object_id = BaseTelepathyObject.__object_counter

    def __eq__(self, other):
        """
        Enables "if conn in connections:".
        WARNING: returns true for connection objects as well, even though
        they behave differently in many cases.
        """
        if other is None:
            return False
        if (self.requested_bus_name == other.requested_bus_name and
                self.object_path == other.object_path ):
            return True
        else:
            return False

    def call_when_ready(self, handler):
        if isinstance(self, ReadyableObject):
            return ReadyableObject.call_when_ready(self, handler)
        else:
            self.call_when_base_ready(handler)

    def is_ready(self):
        if isinstance(self, ReadyableObject):
            return ReadyableObject.is_ready(self)
        else:
            return self.is_base_ready()

    def call_when_base_ready(self, handler):
        if self.is_base_ready():
            do_reply(handler)
        else:
            self._base_ready_handlers.append(handler)

    def is_base_ready(self):
        return self.__base_ready and InterfaceBuilder.is_base_ready(self)

    def base_ready(self):
        if self.__base_ready is True:
            return

        self.__base_ready = True
        InterfaceBuilder.base_ready(self, self._interface_builder_ready_cb)

    def _interface_builder_ready_cb(self):
        self._register_callbacks()
        ready = getattr(self, 'BaseReady', None)
        if ready is not None:
            self.BaseReady()
        while(len(self._base_ready_handlers) > 0):
            handler = self._base_ready_handlers.pop(0)
            do_reply(handler)

    def __repr__ (self):
        return "<%s %s>" % (self.__class__.__name__, self.__object_id)


    def get_related_object(self, klass, object_path):
        """
        Quite often, object paths are returned as strings. This will convert
        them to objects that you can actually use (InterfaceFactory objects).
        """
        return klass(self.requested_bus_name, object_path, self._bus)

class BaseAccountManager(BaseTelepathyObject):
    """
    Listens to an Account manager, and prints unhandled events.
    This is the base class for everything that deals with remote Account
    manager objects. Classes should derive from this to add functionality.
    It is assumed that each piece of functionality will be in its own class,
    with a single class to multiple-inheret everything. For this reason, it is
    recommended that you try to avoid name collisions in sister-classes. This
    can be done reliably using the __ prefix for members, and only writing one
    signal handler per signal. If common functionality is needed in sister
    classes, push it up to the parent class.

    WARNING: connections are not guaranteed to be ready when this is initialised.
    This is what finish_setup is for.
    """
    def __init__(self, bus_name = "org.freedesktop.Telepathy.AccountManager",
                 object_path = None, bus = None):
        BaseTelepathyObject.__init__(self, telepathy.interfaces.ACCOUNT_MANAGER,
                                bus_name, object_path, bus)
        logger.info("Listening to Account manager %s"  % self.object_path)
        self.base_ready()

    # FIXME: this should really be auto-generated from the spec globally.
    _signal_names ={
        'org.freedesktop.Telepathy.AccountManager':
            ['AccountRemoved', 'AccountValidityChanged'],
        }
    _method_names ={
        'org.freedesktop.Telepathy.AccountManager':
            ['CreateAccount'],
        }

class BaseAccount(BaseTelepathyObject):
    """
    Listens to an Account, and prints unhandled events.
    This is the base class for everything that deals with remote Account
    objects. Classes should derive from this to add functionality.
    It is assumed that each piece of functionality will be in its own class,
    with a single class to multiple-inheret everything. For this reason, it is
    recommended that you try to avoid name collisions in sister-classes. This
    can be done reliably using the __ prefix for members, and only writing one
    signal handler per signal. If common functionality is needed in sister
    classes, push it up to the parent class.

    WARNING: connections are not guaranteed to be ready when this is initialised.
    This is what finish_setup is for.
    """
    def __init__(self, bus_name = None, object_path = None, bus = None):
        BaseTelepathyObject.__init__(self, telepathy.interfaces.ACCOUNT,
                                bus_name, object_path, bus)
        logger.info("Listening to Account %s" % self.object_path)
        self.base_ready()

    # FIXME: this should really be auto-generated from the spec globally.
    _signal_names ={
        'org.freedesktop.Telepathy.Account':
            ['Removed', 'AccountPropertyChanged'],
        'org.freedesktop.Telepathy.Account.Interface.Avatar':
            ['AvatarChanged'],
        }
    _method_names ={
        'org.freedesktop.Telepathy.Account':
            ['Remove', 'UpdateParameters', 'Reconnect'],
        }

class BaseChannelDispatcher(BaseTelepathyObject):
    """
    Listens to the Channel Dispatch, and prints unhandled events.
    This is the base class for everything that deals with remote Channel
    Dispatcher objects. Classes should derive from this to add functionality.
    It is assumed that each piece of functionality will be in its own class,
    with a single class to multiple-inheret everything. For this reason, it is
    recommended that you try to avoid name collisions in sister-classes. This
    can be done reliably using the __ prefix for members, and only writing one
    signal handler per signal. If common functionality is needed in sister
    classes, push it up to the parent class.

    WARNING: connections are not guaranteed to be ready when this is initialised.
    This is what finish_setup is for.
    """
    def __init__(self, bus_name = "org.freedesktop.Telepathy.AccountManager",
                 object_path = "/org/freedesktop/Telepathy/ChannelDispatcher",
                 bus = None):
        BaseTelepathyObject.__init__(self, telepathy.interfaces.CHANNEL_DISPATCHER,
                                bus_name, object_path, bus)
        logger.info("Listening to Channel Dispatch %s"  % self.object_path)
        self.base_ready()

    # FIXME: this should really be auto-generated from the spec globally.
    _signal_names ={
        'org.freedesktop.Telepathy.ChannelDispatcher':
            [],
        'org.freedesktop.Telepathy.ChannelDispatcher.Interface.OperationList':
            ['NewDispatchOperation', 'DispatchOperationFinished'],
        }
    _method_names ={
        'org.freedesktop.Telepathy.ChannelDispatcher':
            ['CreateChannel', 'EnsureChannel'],
        'org.freedesktop.Telepathy.ChannelDispatcher.Interface.OperationList':
            [],
        }


class BaseChannelRequest(BaseTelepathyObject):
    """
    Listens to a Channel Request, and prints unhandled events.
    This is the base class for everything that deals with remote Channel
    Request objects. Classes should derive from this to add functionality.
    It is assumed that each piece of functionality will be in its own class,
    with a single class to multiple-inheret everything. For this reason, it is
    recommended that you try to avoid name collisions in sister-classes. This
    can be done reliably using the __ prefix for members, and only writing one
    signal handler per signal. If common functionality is needed in sister
    classes, push it up to the parent class.

    WARNING: connections are not guaranteed to be ready when this is initialised.
    This is what finish_setup is for.
    """
    def __init__(self, bus_name = None, object_path = None, bus = None):
        BaseTelepathyObject.__init__(self, telepathy.interfaces.CHANNEL_REQUEST,
                                bus_name, object_path, bus)
        logger.info("Listening to Channel Request %s"  % self.object_path)
        self.base_ready()

    # FIXME: this should really be auto-generated from the spec globally.
    _signal_names ={
        'org.freedesktop.Telepathy.ChannelRequest':
            ['Failed', 'Succeeded'],
        }
    _method_names ={
        'org.freedesktop.Telepathy.ChannelRequest':
            ['Proceed', 'Cancel'],
        }

class BaseConnectionManager(BaseTelepathyObject):
    """
    Listens to a connection manager, and prints unhandled events.
    This is the base class for everything that deals with remote connection
    manager objects. Classes should derive from this to add functionality.
    It is assumed that each piece of functionality will be in its own class,
    with a single class to multiple-inheret everything. For this reason, it is
    recommended that you try to avoid name collisions in sister-classes. This
    can be done reliably using the __ prefix for members, and only writing one
    signal handler per signal. If common functionality is needed in sister
    classes, push it up to the parent class.

    WARNING: connections are not guaranteed to be ready when this is initialised.
    This is what finish_setup is for.
    """
    def __init__(self, bus_name = None, object_path = None, bus = None):
        BaseTelepathyObject.__init__(self, telepathy.interfaces.CONNECTION_MANAGER,
                                bus_name, object_path, bus)
        logger.info("Listening to Connection Manager %s" % self.object_path)
        self.base_ready()

    # FIXME: this should really be auto-generated from the spec globally.
    _signal_names = {
        'org.freedesktop.Telepathy.ConnectionManager':
            ['NewConnection'],
        }
    _method_names = {
        'org.freedesktop.Telepathy.ConnectionManager':
            ['GetParameters', 'ListProtocols', 'RequestConnection'],
        }

class BaseConnection(BaseTelepathyObject):
    """
    Listens to a connection, and prints unhandled events.
    This is the base class for everything that deals with remote connection
    objects. Classes should derive from this to add functionality.
    It is assumed that each piece of functionality will be in its own class,
    with a single class to multiple-inheret everything. For this reason, it is
    recommended that you try to avoid name collisions in sister-classes. This
    can be done reliably using the __ prefix for members, and only writing one
    signal handler per signal. If common functionality is needed in sister
    classes, push it up to the parent class.

    WARNING: connections are not guaranteed to be ready when this is initialised.
    This is what finish_setup is for.
    """
    def __init__(self, bus_name = None, object_path = None, bus = None,
                 error_handler=default_error_handler):
        BaseTelepathyObject.__init__(self, telepathy.interfaces.CONNECTION,
                                bus_name, object_path, bus)
        logger.info("Listening to Connection %s" % self.object_path)
        self._status_changed_connection = \
            self.connect_to_signal('StatusChanged',
                lambda status, reason: self._status_cb(status))
        self.GetStatus(reply_handler=self._status_cb,
                       error_handler=error_handler,
                       ignore_ready=True)

    def _status_cb(self, status):
        from telepathy.constants import CONNECTION_STATUS_CONNECTED
        if status == CONNECTION_STATUS_CONNECTED:
            self.base_ready()

            if self._status_changed_connection:
                # disconnect signal handler
                self._status_changed_connection.remove()
                self._status_changed_connection = None

    # FIXME: this should really be auto-generated from the spec globally.
    _signal_names = {
        'org.freedesktop.Telepathy.Connection':
            ['SelfHandleChanged', 'NewChannel',
             'ConnectionError', 'StatusChanged'],
        'org.freedesktop.Telepathy.Connection.Interface.Aliasing':
            ['AliasesChanged'],
        'org.freedesktop.Telepathy.Connection.Interface.Avatars':
            ['AvatarUpdated', 'AvatarRetrieved'],
        'org.freedesktop.Telepathy.Connection.Interface.Capabilities':
            ['CapabilitiesChanged'],
        'org.freedesktop.Telepathy.Connection.Interface.ContactCapabilities':
            ['ContactCapabilitiesChanged'],
        # Unstable
        #'org.freedesktop.Telepathy.Connection.Interface.ContactInfo.DRAFT':
        #    ['ContactInfoChanged'],
        'org.freedesktop.Telepathy.Connection.Interface.Contacts':
            [],
        'org.freedesktop.Telepathy.Connection.Interface.Location':
            ['LocationUpdated'],
        # Deprecated
        'org.freedesktop.Telepathy.Connection.Interface.Presence':
            ['PresenceUpdate'],
        # Unstable
        #'org.freedesktop.Telepathy.Connection.Interface.Renaming':
        #    ['Renamed'],
        'org.freedesktop.Telepathy.Connection.Interface.Requests':
            ['NewChannels', 'ChannelClosed'],
        'org.freedesktop.Telepathy.Connection.Interface.SimplePresence':
            ['PresencesChanged'],
        }

    _method_names = {
        'org.freedesktop.Telepathy.Connection':
            ['Connect', 'Disconnect', 'GetInterfaces', 'GetProtocol',
             'GetSelfHandle', 'GetStatus', 'HoldHandles', 'InspectHandles',
             'ListChannels', 'ReleaseHandles', 'RequestChannel',
             'RequestHandles' ],
        'org.freedesktop.Telepathy.Connection.Interface.Aliasing':
            ['GetAliasFlags', 'RequestAliases', 'GetAliases', 'SetAliases'],
        'org.freedesktop.Telepathy.Connection.Interface.Avatars':
            ['GetAvatarRequirements', 'GetAvatarTokens', 'GetKnownAvatarTokens',
             'RequestAvatar', 'RequestAvatars', 'SetAvatar', 'ClearAvatar'],
        'org.freedesktop.Telepathy.Connection.Interface.Capabilities':
            ['AdvertiseCapabilities', 'GetCapabilities'],
        'org.freedesktop.Telepathy.Connection.Interface.ContactCapabilities':
            ['UpdateCapabilities', 'GetContactCapabilities'],
        # Unstable
        #'org.freedesktop.Telepathy.Connection.Interface.ContactInfo.DRAFT':
        #    ['GetContactInfo', 'RequestContactInfo', 'SetContactInfo'],
        'org.freedesktop.Telepathy.Connection.Interface.Contacts':
            ['GetContactAttributes'],
        'org.freedesktop.Telepathy.Connection.Interface.Location':
            ['GetLocations', 'RequestLocation', 'SetLocation'],
        # Deprecated
        'org.freedesktop.Telepathy.Connection.Interface.Presence':
            ['AddStatus', 'ClearStatus', 'GetPresence', 'GetStatuses',
             'RemoveStatus', 'RequestPresence', 'SetLastActivityTime',
             'SetStatus'],
        # Unstable
        #'org.freedesktop.Telepathy.Connection.Interface.Renaming':
        #    ['RequestRename'],
        'org.freedesktop.Telepathy.Connection.Interface.Requests':
            ['CreateChannel', 'EnsureChannel'],
        'org.freedesktop.Telepathy.Connection.Interface.SimplePresence':
            ['SetPresence', 'GetPresences'],
        }

class BaseChannel(BaseTelepathyObject):
    """
    The base class for all things which listen/respond to channels.
    The method for registering channel handlers with subclasses of
    ChannelDispatcher is currently in flux. In the past, you used the
    ChannelDispatcher.handler_for decorator, which is made of bong.
    What you now need to do is ensure that your ChannelListener subclass has
    non-None values for channel_type, handle_type, and capability_flags, and
    then ensure that your ChannelDispatcher subclass/mixin contains a dict.

    The channel object will be ready for use when the constructor is called.
    """

    def __init__(self, bus_name = None, object_path = None, bus = None):
        BaseTelepathyObject.__init__(self, telepathy.interfaces.CHANNEL,
                                bus_name, object_path, bus)
        logger.info("Listening to Channel %s" % self.object_path)
        self.base_ready()

    _signal_names = {
        'org.freedesktop.Telepathy.Channel':
            ['Closed'],
        # Unstable
        #'org.freedesktop.Telepathy.Channel.FUTURE':
        #    [],

        # Channel types
        'org.freedesktop.Telepathy.Channel.Type.ContactList':
            [],
        'org.freedesktop.Telepathy.Channel.Type.StreamedMedia':
            ['StreamAdded', 'StreamDirectionChanged', 'StreamError',
            'StreamRemoved', 'StreamStateChanged'],
        'org.freedesktop.Telepathy.Channel.Type.RoomList':
            ['GotRooms', 'ListingRooms'],
        'org.freedesktop.Telepathy.Channel.Type.Text':
            ['LostMessage', 'Received', 'SendError', 'Sent'],
        # Deprecated
        #'org.freedesktop.Telepathy.Channel.Type.Tubes':
        #    ['NewTube', 'TubeStateChanged', 'TubeClosed',
        #     'DBusNamesChanged', 'StreamTubeNewConnection'],
        'org.freedesktop.Telepathy.Channel.Type.StreamTube':
            ['NewRemoteConnection', 'NewLocalConnection', 'ConnectionClosed'],
        'org.freedesktop.Telepathy.Channel.Type.DBusTube':
            ['DBusNamesChanged'],
        'org.freedesktop.Telepathy.Channel.Type.FileTransfer':
            ['FileTransferStateChanged', 'TransferredBytesChanged',
            'InitialOffsetDefined'],
        # Unstable
        #'org.freedesktop.Telepathy.Channel.Type.ContactSearch.DRAFT2':
        #    ['SearchStateChanged', 'SearchResultReceived'],

        # Channel interfaces
        # Unstable
        #'org.freedesktop.Telepathy.Channel.Interface.CallMerging':
        #    [],
        'org.freedesktop.Telepathy.Channel.Interface.CallState':
            ['CallStateChanged'],
        'org.freedesktop.Telepathy.Channel.Interface.ChatState':
            ['ChatStateChanged'],
        'org.freedesktop.Telepathy.Channel.Interface.Destroyable':
            [],
        'org.freedesktop.Telepathy.Channel.Interface.DTMF':
            [],
        'org.freedesktop.Telepathy.Channel.Interface.Group':
            ['HandleOwnersChanged', 'SelfHandleChanged', 'GroupFlagsChanged',
             'MembersChanged', 'MembersChangedDetailed'],
        'org.freedesktop.Telepathy.Channel.Interface.Hold':
            ['HoldStateChanged'],
        # Unstable
        #'org.freedesktop.Telepathy.Channel.Interface.HTML.DRAFT':
        #    [],
        'org.freedesktop.Telepathy.Channel.Interface.Password':
            ['PasswordFlagsChanged'],
        'org.freedesktop.Telepathy.Channel.Interface.MediaSignalling':
            ['NewSessionHandler'],
        'org.freedesktop.Telepathy.Channel.Interface.Messages':
            ['MessageSent', 'PendingMessagesRemoved', 'MessageReceived'],
        'org.freedesktop.Telepathy.Channel.Interface.Tube':
            ['TubeChannelStateChanged'],
        'org.freedesktop.Telepathy.Properties':
            ['PropertiesChanged', 'PropertyFlagsChanged'],
        }
    _method_names = {
        'org.freedesktop.Telepathy.Channel':
            ['Close', 'GetChannelType', 'GetHandle', 'GetInterfaces'],
        # Unstable
        #'org.freedesktop.Telepathy.Channel.FUTURE':
        #    [],

        # Channel types
        'org.freedesktop.Telepathy.Channel.Type.ContactList':
            [],
        'org.freedesktop.Telepathy.Channel.Type.StreamedMedia':
            ['ListStreams', 'RemoveStreams', 'RequestStreamDirection',
             'RequestStreams'],
        'org.freedesktop.Telepathy.Channel.Type.RoomList':
            ['GetListingRooms', 'ListRooms', 'StopListing'],
        'org.freedesktop.Telepathy.Channel.Type.Text':
            ['AcknowledgePendingMessages', 'GetMessageTypes',
             'ListPendingMessages', 'Send'],
        # Deprecated
        #'org.freedesktop.Telepathy.Channel.Type.Tubes':
        #    ['GetAvailableStreamTubeTypes', 'GetAvailableTubeTypes',
        #     'ListTubes', 'OfferDBusTube', 'OfferStreamTube', 'AcceptDBusTube',
        #     'AcceptStreamTube', 'CloseTube', 'GetDBusTubeAddress',
        #     'GetDBusNames', 'GetStreamTubeSocketAddress'],
        'org.freedesktop.Telepathy.Channel.Type.StreamTube':
            ['Offer', 'Accept'],
        'org.freedesktop.Telepathy.Channel.Type.DBusTube':
            ['Offer', 'Accept'],
        'org.freedesktop.Telepathy.Channel.Type.FileTransfer':
            ['AcceptFile', 'ProvideFile'],
        # Unstable
        #'org.freedesktop.Telepathy.Channel.Type.ContactSearch.DRAFT2':
        #    ['Search', 'More', 'Stop'],

        # Channel interfaces
        # Unstable
        #'org.freedesktop.Telepathy.Channel.Interface.CallMerging':
        #    ['Merge', 'Split'],
        'org.freedesktop.Telepathy.Channel.Interface.CallState':
            ['GetCallStates'],
        'org.freedesktop.Telepathy.Channel.Interface.ChatState':
            ['SetChatState'],
        'org.freedesktop.Telepathy.Channel.Interface.Destroyable':
            ['Destroy'],
        'org.freedesktop.Telepathy.Channel.Interface.DTMF':
            ['StartTone', 'StopTone'],
        'org.freedesktop.Telepathy.Channel.Interface.Group':
            ['AddMembers', 'GetAllMembers', 'GetGroupFlags', 'GetHandleOwners',
             'GetLocalPendingMembers', 'GetLocalPendingMembersWithInfo',
             'GetMembers', 'GetRemotePendingMembers', 'GetSelfHandle',
             'RemoveMembers', 'RemoveMembersWithReason'],
        'org.freedesktop.Telepathy.Channel.Interface.Hold':
            ['GetHoldState', 'RequestHold'],
        # Unstable
        #'org.freedesktop.Telepathy.Channel.Interface.HTML.DRAFT':
        #    [],
        'org.freedesktop.Telepathy.Channel.Interface.Password':
            ['GetPasswordFlags', 'ProvidePassword'],
        'org.freedesktop.Telepathy.Channel.Interface.MediaSignalling':
            ['GetSessionHandlers'],
        'org.freedesktop.Telepathy.Channel.Interface.Messages':
            ['SendMessage', 'GetPendingMessageContent'],
        'org.freedesktop.Telepathy.Channel.Interface.Tube':
            [],
        'org.freedesktop.Telepathy.Properties':
            ['GetProperties', 'ListProperties', 'SetProperties'],
        }
