# -.- coding: utf-8 -.-

# Teamgeist
#
# Copyright © 2009 Collabora Limited
# * Contact: Youness Alaoui
# Copyright © 2009 Siegfried-Angel Gevatter Pujals <rainct@ubuntu.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# * Contributors:
# *   Youness Alaoui, Collabora Ltd.
# *   Siegfried-Angel Gevatter Pujals


import sqlite3
import time
from teamgeist import RemoteEvent, EventSubject

class LookupTable(object):
    def __init__(self, table):
        self._table = table
        self._id_to_value = {}
        self._value_to_id = {}
    def add(self, id, value):
        self._value_to_id[value] = id
        self._id_to_value[id] = value
    def id(self, value):
        return self._value_to_id.get(value, None)
    def value(self, id):
        return self._id_to_value.get(id, None)

class DBAccessor(object):
    """
    This class is meant to provide an easy to use API to access Teamgeist's
    database.
    """

    # if owner is NULL, then the UNIQUE constraint is not checked
    # anymore. so we must define a unique owner uid for the local events
    LOCAL_EVENT_ID = ':self:'

    def __init__(self, connection = None):
        if connection:
            # TODO: do we really want to use zg db ? what about storage state?
            self._connection = connection
        else:
            # TODO: fix path to db
            self._connection = sqlite3.connect("./teamgeist.sqlite")
        self._cursor = self._connection.cursor()
        pragma = self._cursor.execute("PRAGMA foreign_keys").fetchall()

        # TODO: add support for libsql without foreign keys support
        if pragma:
            self._supports_foreign_keys = True
        else:
            self._supports_foreign_keys = False

        self._lookup_objs = {}

        self._create_tables()

    def _create_tables(self):
        # TODO: what's the diff between primary key and unique? how to properly
        # design this thing??
        self._cursor.executescript("""
                PRAGMA foreign_keys = ON;

		CREATE TABLE IF NOT EXISTS interpretation
			(id INTEGER PRIMARY KEY, value VARCHAR UNIQUE);
		CREATE UNIQUE INDEX IF NOT EXISTS interpretation_value
			ON interpretation(value);

		CREATE TABLE IF NOT EXISTS manifestation
			(id INTEGER PRIMARY KEY, value VARCHAR UNIQUE);
		CREATE UNIQUE INDEX IF NOT EXISTS manifestation_value
			ON manifestation(value);

		CREATE TABLE IF NOT EXISTS actor
			(id INTEGER PRIMARY KEY, value VARCHAR UNIQUE);
		CREATE UNIQUE INDEX IF NOT EXISTS actor_value
			ON actor(value);

		CREATE TABLE IF NOT EXISTS payload
			(id INTEGER PRIMARY KEY, value BLOB UNIQUE);

		CREATE TABLE IF NOT EXISTS uri
			(id INTEGER PRIMARY KEY, value VARCHAR UNIQUE);
		CREATE UNIQUE INDEX IF NOT EXISTS uri_value ON uri(value);

		CREATE TABLE IF NOT EXISTS mimetype
			(id INTEGER PRIMARY KEY, value VARCHAR UNIQUE);
		CREATE UNIQUE INDEX IF NOT EXISTS mimetype_value
			ON mimetype(value);

		CREATE TABLE IF NOT EXISTS text
			(id INTEGER PRIMARY KEY, value VARCHAR UNIQUE);
		CREATE UNIQUE INDEX IF NOT EXISTS text_value
			ON text(value);


		CREATE TABLE IF NOT EXISTS storage
			(id INTEGER PRIMARY KEY,
			 value VARCHAR UNIQUE);
		CREATE UNIQUE INDEX IF NOT EXISTS storage_value
			ON storage(value);


            CREATE TABLE IF NOT EXISTS Teams (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                type VARCHAR NOT NULL,
                account_id VARCHAR NOT NULL,
                uid VARCHAR NOT NULL,
                joinable BOOLEAN NOT NULL,
                CONSTRAINT unique_team UNIQUE (type, account_id, uid));

            CREATE TABLE IF NOT EXISTS SharedEvents (
                team_id INTEGER NOT NULL
                     REFERENCES Teams (id) ON DELETE CASCADE,
                owner VARCHAR DEFAULT NULL,
                remote_id INTEGER NOT NULL,
                added_timestamp INTEGER NOT NULL,
                timestamp INTEGER DEFAULT NULL,
                interpretation INTEGER DEFAULT NULL
                     REFERENCES interpretation (id) ON DELETE CASCADE,
                manifestation INTEGER DEFAULT NULL
                     REFERENCES manifestation (id) ON DELETE CASCADE,
                actor INTEGER DEFAULT NULL
                     REFERENCES actor (id) ON DELETE CASCADE,
                payload INTEGER DEFAULT NULL
                     REFERENCES payload (id) ON DELETE CASCADE,
                subj_id INTEGER NOT NULL
                     REFERENCES uri (id) ON DELETE CASCADE,
                subj_interpretation INTEGER DEFAULT NULL
                     REFERENCES interpretation (id) ON DELETE CASCADE,
                subj_manifestation INTEGER DEFAULT NULL
                     REFERENCES manifestation (id) ON DELETE CASCADE,
                subj_origin INTEGER DEFAULT NULL
                     REFERENCES uri (id) ON DELETE CASCADE,
                subj_mimetype INTEGER DEFAULT NULL
                     REFERENCES mimetype (id) ON DELETE CASCADE,
                subj_text INTEGER DEFAULT NULL
                     REFERENCES text (id) ON DELETE CASCADE,
                subj_storage INTEGER DEFAULT NULL
                     REFERENCES storage (id) ON DELETE CASCADE,
                CONSTRAINT unique_event UNIQUE (team_id, owner,
                                                remote_id, subj_id),
                CONSTRAINT unique_event2 UNIQUE (team_id, owner, timestamp,
                                                interpretation, manifestation,
                                                actor, subj_id));

            CREATE TABLE IF NOT EXISTS Topics (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                name VARCHAR NOT NULL,
                team_id INTEGER NOT NULL
                     REFERENCES Teams (id) ON DELETE CASCADE,
                CONSTRAINT unique_topics UNIQUE (name, team_id));

            CREATE TABLE IF NOT EXISTS Search_criteria (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                topic_id INTEGER NOT NULL
                     REFERENCES Topics (id) ON DELETE CASCADE,
                criteria VARCHAR NOT NULL,
                match INTEGER NOT NULL,
                value VARCHAR NOT NULL);

            """)
        self._connection.commit()


    def GetTopics(self, team):
        """
        Return all the topics from the database.
        If team is not None, return only the topics for that team
        Returns a list Teamgeist.Topics
        """
        result = self._cursor.execute(
            """SELECT t.id, t.name,
                      s.id AS search_id, s.criteria, s.match, s.value
                        FROM Topics AS t, Search_criteria AS s
                          WHERE t.team_id = ? AND t.id = s.topic_id
                             ORDER BY t.id""",
            (team.id,)).fetchall() or []
        topics = []
        topic_names = {}
        search_criterias = {}
        for row in result:
            search_criterias.setdefault(row[0], []).append(row[2:])
            topic_names[row[0]] = row[1]
        for topic in search_criterias.keys():
            topics.append((topic, topic_names[topic], search_criterias[topic]))

        return topics

    def AddTopic(self, team, topic):
        """
        Adds a topic to the database, adds the search criterias, links them to
        the topic and links the topic to the team
        Returns its id
        """

        # TODO: return errors for stuff like duplicate topic name, etc..
        if len(topic.search_criterias) == 0:
            return 0
        self._cursor.execute(
            "INSERT INTO Topics (id, name, team_id) VALUES (NULL, ?, ?)",
            (topic.name, team.id,))
        topic_id = self._cursor.lastrowid

        result = self._cursor.execute(
            "SELECT MAX(id) FROM Search_criteria").fetchone() or (0,)

        search_id = result[0] or 0
        args = []
        for search in topic.search_criterias:
            search_id += 1
            search.id = search_id
            args.extend((search_id, topic_id, search.criteria,
                         search.match, search.value))
        self._cursor.execute("""
            INSERT INTO Search_criteria (id, topic_id, criteria, match, value)
            %s """ % " UNION ALL ".join(
            ["SELECT ?, ?, ?, ?, ?"] * len(topic.search_criterias)), args)

        self._connection.commit()

        topic.id = topic_id
        return topic_id

    def DeleteTopic(self, topic):
        """
        Deletes a topic from the database as well as its search criterias and its
        association with the team
        """
        # TODO: support self._supports_foreign_keys == False
        # Thanks to foreign key references and ON DELETE CASCADE,
        # search criterias will also get deleted
        self._cursor.execute("DELETE FROM Topics WHERE id=?", (topic.id,))
        self._connection.commit()

    def GetTeams(self):
        """
        Returns all a list of all the teams in the database
        Returns a list of Teams
        """
        result = self._cursor.execute(
            "SELECT id, account_id, type, uid, joinable FROM Teams")
        return result.fetchall() or []

    def AddTeam(self, account, type, uid, joinable):
        """
        Adds a Team to the database
        Return the id of the new team
        """
        try:
            self._cursor.execute(
                """INSERT INTO Teams (id, type, account_id, uid, joinable)
                      VALUES (NULL, ?, ?, ?, ?)""",
                (type, account, uid, int(joinable)))
        except sqlite3.IntegrityError, e:
            return 0
        self._connection.commit()
        return self._cursor.lastrowid

    def DeleteTeam(self, team):
        """
        Deletes a team from the database as well as its associated
        topics, search criterias, events
        """
        # TODO: support self._supports_foreign_keys == False
        # Thanks to foreign key references and ON DELETE CASCADE topics,
        # search criterias, events will also get deleted
        self._cursor.execute("DELETE FROM Teams WHERE id=?", (team.id,))
        self._connection.commit()

    def SetTeamJoinable(self, team, joinable):
        """
        Changes whether a team should be joinable or not
        """
        self._cursor.execute("UPDATE Teams SET joinable=? WHERE id=?",
            (bool(joinable), team.id))
        self._connection.commit()

    def _ensure_lkp_id(self, table, value):
        if not table in self._lookup_objs:
            self._lookup_objs[table] = LookupTable(table)
            rows = self._cursor.execute(
                "SELECT id, value FROM %s" % table).fetchall()
            for row in rows:
                if isinstance(row[1], sqlite3.Binary):
                    self._lookup_objs[table].add(row[0], "".join(row[1]))
                else:
                    self._lookup_objs[table].add(row[0], row[1])

        lkp = self._lookup_objs[table]
        id = lkp.id(value)
        if id is not None:
            return id
        else:
            if table == 'payload':
                insert_val = sqlite3.Binary(value)
            else:
                insert_val = value
            self._cursor.execute(
                "INSERT OR IGNORE INTO %s (value) VALUES (?)" % table,
                (insert_val,))
            lkp.add(self._cursor.lastrowid, value)
            return self._cursor.lastrowid

    def _lookup_value(self, table, id):
        if not table in self._lookup_objs:
            self._lookup_objs[table] = LookupTable(table)
            rows = self._cursor.execute(
                "SELECT id, value FROM %s" % table).fetchall()
            for row in rows:
                self._lookup_objs[table].add(row[0], row[1])

        return self._lookup_objs[table].value(id)


    def InsertEvent (self, team, event):
        """
        Insert an event into the database
        If the owner is NULL, then only insert the id and subj_id
        and let the rest be taken from the local event database
        using the remote_id attribute as the local id
        """
        timestamp = int(time.time() * 1000)

        # TODO: Check for constraint error
        # TODO: bulk add multiple uris/whatever ?

        args = []
        for subject in event.subjects:
            interpretation = self._ensure_lkp_id('interpretation',
                                                 event.interpretation)
            manifestation = self._ensure_lkp_id('manifestation',
                                                event.manifestation)
            actor = self._ensure_lkp_id('actor', event.actor)
            payload = self._ensure_lkp_id('payload', event.payload)
            try:
                subject_id = self._ensure_lkp_id('uri', subject.id)
            except:
                subject_id = self._ensure_lkp_id('uri', subject.uri)
            subject_interpretation = self._ensure_lkp_id('interpretation',
                                                         subject.interpretation)
            subject_manifestation = self._ensure_lkp_id('manifestation',
                                                        subject.manifestation)
            subject_origin = self._ensure_lkp_id('uri', subject.origin)
            subject_mimetype = self._ensure_lkp_id('mimetype', subject.mimetype)
            subject_text = self._ensure_lkp_id('text', subject.text)
            subject_storage = self._ensure_lkp_id('storage', subject.storage)

            args.extend((team.id, event.owner or self.LOCAL_EVENT_ID, event.id,
                         timestamp, event.timestamp, interpretation,
                         manifestation, actor, payload,
                         subject_id, subject_interpretation,
                         subject_manifestation, subject_origin,
                         subject_mimetype, subject_text, subject_storage))

        sql = """INSERT OR IGNORE INTO SharedEvents (team_id, owner, remote_id,
                    added_timestamp, timestamp, interpretation, manifestation,
                    actor, payload, subj_id, subj_interpretation,
                    subj_manifestation, subj_origin, subj_mimetype,
                    subj_text, subj_storage)
                     %s """ % " UNION ALL ".join(
                ["SELECT " + 15 * "?, " + "?"] * len(event.subjects))
        print args
        self._cursor.execute(sql, args)
        event.added_timestamp = timestamp
        self._connection.commit()

    def GetEvents (self, team, owner, event_ids):
        """
        Get the events with the ids requested.
        If the owner is NULL, then fetch the id from the local event database
        using the remote_id attribute as the local id
        """
        rows = self._cursor.execute("""
                      SELECT owner, remote_id, added_timestamp,
                             timestamp, interpretation, manifestation,
                             actor, payload, subj_id, subj_interpretation,
                             subj_manifestation, subj_origin, subj_mimetype,
                             subj_text, subj_storage
                        FROM SharedEvents
                           WHERE team_id = ? AND owner = ? AND remote_id IN (%s)
                      """ % ",".join(["?" for id in event_ids]),
                                     (team.id, owner or self.LOCAL_EVENT_ID,) + \
                                         tuple(event_ids)).fetchall()

        events = {}
        for row in rows:
            event_id = row[1]

            subject_id = self._lookup_value('uri', row[8])
            subj_interpretation = self._lookup_value('interpretation', row[9])
            subj_manifestation = self._lookup_value('manifestation', row[10])
            subj_origin = self._lookup_value('uri', row[11])
            subj_mimetype = self._lookup_value('mimetype', row[12])
            subj_text = self._lookup_value('text', row[13])
            subj_storage = self._lookup_value('storage', row[14])

            subject = EventSubject(event_id,
                                   subject_id,
                                   subj_interpretation,
                                   subj_manifestation,
                                   subj_origin,
                                   subj_mimetype,
                                   subj_text,
                                   subj_storage)
            if event_id in events:
                events[event_id].subjects.append(subject)
            else:
                interpretation = self._lookup_value('interpretation', row[4])
                manifestation = self._lookup_value('manifestation', row[5])
                actor = self._lookup_value('actor', row[6])
                payload = self._lookup_value('payload', row[7])
                events[event_id] = RemoteEvent(team,
                                               row[0],
                                               row[2],
                                               event_id,
                                               row[3],
                                               interpretation,
                                               manifestation,
                                               actor,
                                               payload, [subject,])

        # Sort events into the requested order
        sorted_events = []
        for id in event_ids:
            # if we are not able to get an event by the given id
            # append None instead of raising an Error
            sorted_events.append(events.get(id, None))

        return sorted_events

    def FindEventIds(self, team, time_range, event_templates,
                     max_events, order):
        """
        Search for events using an event template
        Returns a list of (owner, id)
        """
        where = WhereClause(WhereClause.AND)
        if time_range[0] > 0:
            where.add("timestamp >= ?", time_range[0])
        if time_range[1] > 0:
            where.add("timestamp <= ?", time_range[1])

        # TODO: handle the event_templates
        # where.extend(self._build_sql_from_event_templates(templates))

        if not where.may_have_results():
            return []

        sql = """SELECT DISTINCT owner, remote_id FROM SharedEvents
                     WHERE team_id = ?"""
        if where:
            sql += " AND " + where.sql

        sql += (" ORDER BY timestamp DESC",
                " ORDER BY timestamp ASC",
                " GROUP BY subj_uri ORDER BY timestamp DESC",
                " GROUP BY subj_uri ORDER BY timestamp ASC",
                " GROUP BY subj_uri ORDER BY COUNT(id) DESC, timestamp DESC",
                " GROUP BY subj_uri ORDER BY COUNT(id) ASC, timestamp ASC")[order]

        if max_events > 0:
            sql += " LIMIT %d" % max_events

        rows = self._cursor.execute(sql, (team.id,) + \
                                        tuple(where.arguments)).fetchall()

        result = {}
        for row in rows:
            result.setdefault(row[0], []).append(row[1])

        return result

    def GetLocalEventIdsSinceEvent(self, team, event_id):
        result = self._cursor.execute("""
                 SELECT remote_id FROM SharedEvents
                    WHERE team_id = ? AND owner = ? AND
                          remote_id <> ? AND
                          added_timestamp >= (SELECT added_timestamp
                                                 FROM SharedEvents
                                                    WHERE team_id = ? AND
                                                          owner = ? AND
                                                          remote_id = ?)
                 """, (team.id, self.LOCAL_EVENT_ID, event_id,
                       team.id, self.LOCAL_EVENT_ID, event_id,)).fetchall() or []
        def first_element(lst):
            return lst[0]
        return map(first_element, result)

    def GetLastAddedEventForUser(self, team, owner):
        # TODO: what happens if there are two events with same timestamp
        # and we got only one, we might loose the second one!
        id = self._cursor.execute("""
                 SELECT remote_id FROM SharedEvents
                    WHERE team_id = ? AND owner = ? AND
                          added_timestamp = (SELECT MAX(added_timestamp)
                                               FROM SharedEvents
                                                 WHERE team_id = ? AND owner = ?)
                 """, (team.id, owner, team.id, owner,)).fetchone() or (0,)
        return id[0]


class WhereClause:
    AND = " AND "
    OR = " OR "

    def __init__(self, relation):
        self._conditions = []
        self.arguments = []
        self._relation = relation
        self._no_result_member = False

    def __len__(self):
        return len(self._conditions)

    def add(self, condition, arguments):
        if not condition:
            return
        self._conditions.append(condition)
        if not hasattr(arguments, "__iter__"):
            self.arguments.append(arguments)
        else:
            self.arguments.extend(arguments)

    def extend(self, where):
        self.add(where.sql, where.arguments)
        if not where.may_have_results():
            if self._relation == self.AND:
                self.clear()
            self.register_no_result()

    @property
    def sql(self):
        # Do not return "()" if there are no conditions
        if len(self._conditions) > 0:
            return "(" + self._relation.join(self._conditions) + ")"
        else:
            return ""

    def register_no_result(self):
        self._no_result_member = True

    def may_have_results(self):
        """
        Return False if we know from our cached data that the query
        will give no results.
        """
        return len(self._conditions) > 0 or not self._no_result_member

    def clear(self):
        """
        Reset this WhereClause to the state of a newly created one.
        """
        self._conditions = []
        self.arguments = []
        self._no_result_member = False
