# -.- coding: utf-8 -.-

# Teamgeist
#
# Copyright (C) 2009 Collabora Limited
# *  Contact: Youness Alaoui
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# * Contributors:
# *   Youness Alaoui, Collabora Ltd.


import re
import dbus
import dbus.service
from dbus_service import *

from teamgeist import *

class GetPresencesAsync(object):
    def __init__(self, account, contacts, success, error):
        self.account = account
        self.contacts = contacts
        self._success = success
        self._error = error

        account.GetConnection(self.got_connection, self.got_error)

    def got_error(self, error):
        self._error(Exception(error))

    def got_connection(self, connection):
        if connection:
            connection.request_handles(self.contacts,
                                       (self.got_handles, connection),
                                       self.got_error)
        else:
            self._error("Account is not connected")

    def got_handles(self, connection, handles):
        c = Contacts(connection, handles)
        c.GetPresences(self.got_presences, self.got_error)

    def got_presences(self, presences):
        self._success(presences)

class GetAliasesAsync(object):
    def __init__(self, account, contacts, success, error):
        self.account = account
        self.contacts = contacts
        self._success = success
        self._error = error

        account.GetConnection(self.got_connection, self.got_error)

    def got_error(self, error):
        self._error(Exception(error))

    def got_connection(self, connection):
        if connection:
            connection.request_handles(self.contacts,
                                       (self.got_handles, connection),
                                       self.got_error)
        else:
            self._error("Account is not connected")

    def got_handles(self, connection, handles):
        c = Contacts(connection, handles)
        c.GetAliases(self.got_aliases, self.got_error)

    def got_aliases(self, aliases):
        self._success(aliases)


class GetContactInfoAsync(object):
    def __init__(self, contacts, success, error):
        self.presences = None
        self.uids = None
        self.aliases = None
        self.replied = False
        self.contacts = contacts
        self._success = success
        self._error = error

        contacts.GetUIDs(self.got_uids,
                         self.got_error)
        contacts.GetAliases(self.got_aliases,
                            self.got_error)
        contacts.GetPresences(self.got_presences,
                              self.got_error)

    def got_error(self, error):
        self.replied = True
        self._error(Exception(error))

    def got_uids(self, uids):
        self.uids = uids
        self.got_something()

    def got_aliases(self, aliases):
        self.aliases = aliases
        self.got_something()

    def got_presences(self, presences):
        self.presences = presences
        self.got_something()

    def got_something(self):
        if self.replied is False and \
                self.uids is not None and \
                self.aliases is not None and \
                self.presences is not None:
            if len(self.uids) == \
                    len(self.aliases) == len(self.presences):
                self.replied = True
                result = []
                for i in range(len(self.uids)):
                    contact = (self.uids[i],
                               self.aliases[i],
                               self.presences[i])
                    result.append(contact)
                self._success(result)
            else:
                self._error(Exception("Invalid result"))

class TeamObject(dbus.service.Object,
                 TeamInterface,
                 TopicInterface,
                 MemberInterface):
    """
    This class exports a Team object over Dbus.
    """

    def __init__(self, team, teamgeist_service, object_path):
        dbus.service.Object.__init__(self, teamgeist_service._name.get_bus(),
                                     object_path, teamgeist_service._name)

        self.teamgeist_service = teamgeist_service
        self.teamgeist = teamgeist_service.teamgeist
        self.team = team

        def participation_changed(emitter, participation):
            self.ParticipationChanged(participation)
        self.team.connect("ParticipationChanged", participation_changed)

        def event_shared(emitter, event):
            self.EventShared(event.id)
        self.team.connect("EventShared", event_shared)

        def event_received(emitter, owner, event):
            def got_uid(uid):
                self.EventReceived(uid, event.struct)
            def got_error(error):
                print "Got error retreiving owner uid : ", error
                # TODO: how to handle this error?

            owner.GetUID(got_uid, got_error)
        self.team.connect("EventReceived", event_received)


        def member_joined(emitter, member):
            def got_uid(uid):
                self.MemberJoined(uid)
            def got_error(error):
                print "Got error retreiving owner uid : ", error
                # TODO: how to handle this error?

            member.GetUID(got_uid, got_error)
        self.team.connect("MemberJoined", member_joined)

        def member_left(emitter, member):
            def got_uid(uid):
                self.MemberLeft(uid)
            def got_error(error):
                print "Got error retreiving owner uid : ", error
                # TODO: how to handle this error?

            member.GetUID(got_uid, got_error)
        self.team.connect("MemberLeft", member_left)


    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='', out_signature='s',
                         async_callbacks=('_success', '_error'))
    def GetName(self, _success, _error):
        """
        Get the name of the team. This is equivalent to the MUC's name
        """
        def got_error(error):
            _error(Exception(error))

        def got_reply(name):
            _success(name)

        self.team.GetName(got_reply, got_error)

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='s', out_signature='',
                         async_callbacks=('_success', '_error'))
    def SetName(self, name, _success, _error):
        """
        Set the name of the team. This is equivalent to the MUC's name
        """
        def got_error(error):
            _error(Exception(error))

        def got_reply():
            _success()

        self.team.SetName(name, got_reply, got_error)

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='', out_signature='s',
                         async_callbacks=('_success', '_error'))
    def GetSubject(self, _success, _error):
        """
        Get the name of the team. This is equivalent to the MUC's subject
        """
        def got_error(error):
            _error(Exception(error))

        def got_reply(subject):
            _success(subject)

        self.team.GetSubject(got_reply, got_error)

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='s', out_signature='',
                         async_callbacks=('_success', '_error'))
    def SetSubject(self, subject, _success, _error):
        """
        Set the name of the team. This is equivalent to the MUC's subject
        """
        def got_error(error):
            _error(Exception(error))

        def got_reply():
            _success()

        self.team.SetSubject(subject, got_reply, got_error)

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='', out_signature='a(ssu)',
                         async_callbacks=('_success', '_error'))
    def GetMembers(self, _success, _error):
        """
        Returns the list of all the members currently in the team
        Members are returned as (uid, alias, presence) structures.
        The unique identifier string and can be used on the
        org.gnome.zeitgeist.Teamgeist.Team.Interfaces.Member interface
        or for method calls like InviteMember, SuggestTopicToMember, etc..
        """
        def got_error(error):
            _error(Exception(error))

        def got_reply(members):
            GetContactInfoAsync(members, _success, _error)

        self.team.GetMembers(got_reply, got_error)

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='s', out_signature='',
                         async_callbacks=('_success', '_error'))
    def InviteMember(self, user, _success, _error):
        """
        Invite a user to become a member of the team.
        If the current user does not have enough priviledges to invite that
        user, then this will send a request to the owner of the team to
        invite that user
        """
        def got_error(error):
            _error(Exception(error))

        def got_reply():
            _success()

        # TODO: transform user into a Member object
        self.team.InviteMember(user, got_reply, got_error)

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='s', out_signature='',
                         async_callbacks=('_success', '_error'))
    def KickMember(self, user, _success, _error):
        """
        Kick a user from the team.
        If the current user does not have enough priviledges to kick that
        user, then this will send a request to the owner of the team to
        kick that user
        """
        def got_error(error):
            _error(Exception(error))

        def got_reply():
            _success()

        # TODO: transform user into a Member object
        self.team.KickMember(user, got_reply, got_error)

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='', out_signature='',
                         async_callbacks=('_success', '_error'))
    def JoinTeam(self, _success, _error):
        """
        Join the team.
        This should set the team as joinable in the database and attempt to join
        it if the account is available and connected
        """
        def got_error(error):
            _error(Exception(error))

        def got_reply():
            _success()

        self.team.JoinTeam(got_reply, got_error)

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='', out_signature='',
                         async_callbacks=('_success', '_error'))
    def LeaveTeam(self, _success, _error):
        """
        Leave the team.
        This should leave the MUC and make the team non-joinable in the database
        without deleting it and without deleting the associated topics
        """
        def got_error(error):
            _error(Exception(error))

        def got_reply():
            _success()

        self.team.LeaveTeam(got_reply, got_error)

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='', out_signature='as',
                         async_callbacks=('_success', '_error'))
    def ListTopics(self, _success, _error):
        """
        List all the topics available for the team
        """
        def got_error(error):
            _error(Exception(error))

        def got_reply(topics):
            result = []
            for topic in topics:
                result.append(topic.name)
            _success(result)

        self.team.ListTopics(got_reply, got_error)

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='sa(suv)', out_signature='s',
                         async_callbacks=('_success', '_error'))
    def AddTopic(self, topic_name, search_criterias, _success, _error):
        """
        Adds a topic to a team with its associated search criterias
        """
        def got_error(error):
            _error(Exception(error))

        def got_reply(topic):
            _success(topic.id)

        self.team.AddTopic(name, search_criterias, got_reply, got_error)

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='s', out_signature='',
                         async_callbacks=('_success', '_error'))
    def DeleteTopic(self, topic_name, _success, _error):
        """
        Deletes a topic from a team
        """
        def got_error(error):
            _error(Exception(error))

        def got_topics(topics):
            def got_reply():
                _success()

            for topic in topics:
                if topic.name == topic_name:
                    self.team.DeleteTopic(topic, got_reply, got_error)
                    return
            _error(Exception("Invalid topic"))

        self.team.ListTopics(got_topics, got_error)

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='ss', out_signature='')
    def SuggestTopicToMember(self, topic, member):
        """
        Suggests to a member of the team a topic for him to share
        """
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='s', out_signature='')
    def SuggestTopicToTeam(self, topic):
        """
        Suggests to all members of the team a topic to be shared
        """
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='ss', out_signature='')
    def RequestDocument(self, member, document):
        """
        Request a document from a member of the team
        """
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='a{sau}', out_signature='a{sa(asaasay)}')
    def GetEvents(self, event_ids):
        """
        Get the events with the ids {owner: [remote_ids]} requested.
        If the owner is the empty string, then it represents the local user
        """
        result = self.team.GetEvents(event_ids)
        for owner in result.keys():
            for i, event in enumerate(result[owner]):
                if event is None:
                    result[owner][i] = ([], [[]], [])
                else:
                    result[owner][i] = event.struct
        return result

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature="(xx)a(sasaasay)uu",
                         out_signature="a{sau}")
    def FindEventIds(self, time_range, event_templates, max_events, order):
        """
        Search for events using an event template and return
        the id {owner: [remote_ids]} of the events.
        If the owner is the empty string, then it represents the local user
        """
        return self.team.FindEventIds(time_range, event_templates,
                                      max_events, order)

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature="", out_signature="u")
    def GetCurrentParticipation(self):
        """ Returns whether or not the user is currently participating
        in the team.
        0: Joining not requested
        1: Participating
        2: Account is offline
        3: Not a member of the team
        4: Joining
        5: Joining failed
        """
        return self.team.GetCurrentParticipation()


    @dbus.service.method(dbus_interface=TEAM_MEMBER_INTERFACE,
                         in_signature='as', out_signature='au',
                         async_callbacks=('_success', '_error'))
    def GetPresences(self, members, _success, _error):
        """
        Returns the presences of the members
        """
        GetPresencesAsync(self.team.account, members, _success, _error)

    @dbus.service.method(dbus_interface=TEAM_MEMBER_INTERFACE,
                         in_signature='as', out_signature='as',
                         async_callbacks=('_success', '_error'))
    def GetAliases(self, members, _success, _error):
        """
        Returns the aliases of the members
        """
        GetAliasesAsync(self.team.account, members, _success, _error)


    @dbus.service.method(dbus_interface=TEAM_TOPIC_INTERFACE,
                         in_signature='as', out_signature='aa(suv)',
                         async_callbacks=('_success', '_error'))
    def ListSearchCriterias(self, topic_names, _success, _error):
        """
        Return all the search criterias for each requested topic
        An empty list of search criterias means the topic was invalid.
        """
        def got_error(error):
            _error(Exception(error))

        def got_topics(topics):
            result = []
            for topic_name in topic_names:
                found = False
                for topic in topics:
                    if topic.name == topic_name:
                        found = True
                        search_criterias = []
                        for search in topic.search_criterias:
                            search_criterias.append((search.criteria,
                                                     search.match,
                                                     search.value))
                        result.append(search_criterias)
                        break
                if found is False:
                    result.append([])
            _success(result)

        self.team.ListTopics(got_topics, got_error)

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='s')
    def NameChanged(self, name):
        """
        Signal sent when the name of the team is changed
        """
        # TODO
        raise NotImplementedError()

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='s')
    def SubjectChanged(self, subject):
        """
        Signal sent when the subject of the team is changed
        """
        # TODO
        raise NotImplementedError()

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='s')
    def MemberJoined(self, member):
        """
        Signal sent when a member joins the team
        """
        pass

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='s')
    def MemberLeft(self, member):
        """
        Signal sent when a member has left the team
        """
        pass

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='s')
    def UserRequestsInvitation(self, member):
        """
        Signal sent when a user requests you to invite him to the team
        """
        raise NotImplementedError()

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='ss')
    def MemberRequestsKick(self, member, to_kick):
        """
        Signal sent when a member requests you to kick someone from
        the team
        """
        raise NotImplementedError()

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='u')
    def EventShared(self, event_id):
        """
        Signal sent when a local event gets shared with the team
        """
        pass

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='sa(suv)')
    def TopicSuggestionReceived(self, member, search_criterias):
        """
        Signal sent when a team member suggests a topic for you to share.
        Accepting a suggested topic just means calling AddTopic
        """
        raise NotImplementedError()

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='ss')
    def DocumentRequestReceived(self, member, document):
        """
        Signal sent when a team member requests a document from you
        Accepting a share document request just means calling AddTopic with the
        appropriate search criteria for sharing that document with the team, or
        creating a new 'single' user team with the member to share the document
        with and calling AddTopic on that new team.
        """
        raise NotImplementedError()

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='u')
    def ParticipationChanged(self, participating):
        """
        Signal sent when Teamgeist participation in a team changes
        (joining or leaving the team)
        """
        pass

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE,
                         signature='s(asaasay)')
    def EventReceived(self, owner, event):
        """
        Signal sent when an event is received from a team member
        """
        pass

    # TODO: add signals for when the list of authorized members changes?
    # needed to know when someone gets kicked... ?



class TeamgeistService(dbus.service.Object,
                       TeamgeistInterface,
                       AccountInterface,
                       ContactInterface,
                       ReadyableObject):
    """
    This class exports the Teamgeist service over Dbus.
    """
    SERVICE_NAME = "org.gnome.zeitgeist.Teamgeist"
    OBJECT_PATH = "/org/gnome/zeitgeist/teamgeist"

    def __init__(self, teamgeist = None):
        bus_name = dbus.service.BusName(self.SERVICE_NAME, bus=dbus.SessionBus(),
                                   do_not_queue=True)
        dbus.service.Object.__init__(self, bus_name, self.OBJECT_PATH)

        ReadyableObject.__init__(self)

        self._teams = {}
        if teamgeist:
            self.teamgeist = teamgeist
        else:
            self.teamgeist = Teamgeist()
        self.teamgeist.ListTeams(self.list_teams_cb, self.list_teams_error)

    def list_teams_error(self, error):
        print error
        self.ready()

    def list_teams_cb(self, teams):
        for team in teams:
            self._get_team_object(team)
        self.teamgeist.connect("NewTeam", self._new_team)
        self.teamgeist.connect("TeamPurged", self._team_purged)
        self.ready()

    def _new_team(self, emitter, team):
        self._get_team_object(team)

    def _team_purged(self, emitter, team):
        if team in self._teams:
            self.TeamPurged(self._teams[team]._object_path)
            self._teams[team].remove_from_connection()
            del self._teams[team]

    def _get_team_object(self, team):
        if team in self._teams:
            return self._teams[team]
        team_obj = TeamObject(team, self, self._get_team_object_path(team))
        self._teams[team] = team_obj
        self.NewTeam(team_obj._object_path)
        return team_obj

    def _get_team_object_path(self, team):
        def _escape_as_identifier(name):
            _BAD = re.compile(r'(?:^[0-9])|(?:[^A-Za-z0-9])')
            if isinstance(name, unicode):
                name = name.encode('utf-8')
                if not name:
                    return '_'
            return _BAD.sub(lambda match: '_%02x' % ord(match.group(0)), name)

        uid = _escape_as_identifier(team.uid)
        team_path_base = "%s/team/%s/%s/" % (self._object_path, team.type, uid)
        team_path = None
        idx = 0
        while team_path is None:
            path = "%s%s" % (team_path_base, idx)
            exists = False
            for t in self._teams.values():
                if t._object_path == path:
                    exists = True
                    break
            if exists:
                idx += 1
            else:
                team_path = path

        return team_path

    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='', out_signature='as',
                         async_callbacks=('_success', '_error'))
    def ListAccounts(self, _success, _error):
        def got_accounts(accounts):
            result = []
            for account in accounts:
                result.append(account.GetUID())
            _success(result)

        def got_error(error):
            _error(Exception(error))

        self.teamgeist.ListAccounts(got_accounts, got_error)


    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='', out_signature='a(ss)')
    def ListTeamTypes(self):
        """
        List all the teams types that are available.
        The team type is needed for adding new teams
        Returns an array of (team_type, description)
        """
        return self.teamgeist.ListTeamTypes()

    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='', out_signature='ao',
                         async_callbacks=('_success', '_error'))
    def ListTeams(self, _success, _error):
        """
        List all the teams that the user has.
        Teams are returned as objects implementing the
        org.gnome.zeitgeist.Teamgeist.Team interface
        """
        def list_teams_ready():
            _success(map(lambda o: o._object_path, self._teams.values()))

        self.call_when_ready(list_teams_ready)


    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='sss', out_signature='o',
                         async_callbacks=('_success', '_error'))
    def AddTeam(self, account, type, uid, _success, _error):
        """
        This adds the team to the Teams database or makes it joinable if already
        there.
        A org.gnome.zeitgeist.Teamgeist.NewTeam signal will be sent when the team
        is added.
        Returns the object path of the Team
        """
        def add_team_ready():
            def add_team_success(team):
                _success(self._teams[team]._object_path)

            def add_team_error(error):
                _error(error)

            self.teamgeist.AddTeam(account, type, uid,
                                   add_team_success, add_team_error)

        self.call_when_ready(add_team_ready)



    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='o', out_signature='',
                         async_callbacks=('_success', '_error'))
    def PurgeTeam(self, team, _success, _error):
        """
        This purge a team from the Teams database deleting all its associated
        topics and events
        A org.gnome.zeitgeist.Teamgeist.TeamPurged signal will be sent when
        the team is deleted.
        """
        def purge_team_ready():
            def purge_team_success():
                _success()

            def purge_team_error(error):
                _error(error)

            for t in self._teams.keys():
                if self._teams[t]._object_path == team:
                    self.teamgeist.PurgeTeam(t, purge_team_success,
                                             purge_team_error)
                    return
            _error(Exception("Team not found"))

        self.call_when_ready(purge_team_ready)


    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='(asaasay)', out_signature='ao')

    def MatchEvent(self, event):
        """
        Match an event with all topics and share that event with the
        appropriate teams if the event matches topics
        Returns the list of teams with which this event got shared
        """
        local_event = RemoteEvent.CreateFromStruct(None, None, event)
        self.teamgeist.MatchEvent(local_event)
        # TODO: make it async and return the list of matched teams
        return []


    @dbus.service.method(dbus_interface=TEAMGEIST_CONTACT_INTERFACE,
                         in_signature='sas', out_signature='au',
                         async_callbacks=('_success', '_error'))
    def GetPresences(self, account, contacts, _success, _error):
        """
        Returns the presences of the contacts
        """
        def got_error(error):
            _error(Exception(error))

        def got_accounts(accounts):
            for acc in accounts:
                if self.account == acc.GetUID():
                    GetPresencesAsync(acc, contacts, _success, _error)
                    return
            _error(Exception("Invalid account name"))

        self.teamgeist.ListAccounts(got_accounts, got_error)

    @dbus.service.method(dbus_interface=TEAMGEIST_CONTACT_INTERFACE,
                         in_signature='sas', out_signature='as',
                         async_callbacks=('_success', '_error'))
    def GetAliases(self, account, contacts, _success, _error):
        """
        Returns the aliases of the contacts
        """
        def got_error(error):
            _error(Exception(error))

        def got_accounts(accounts):
            for acc in accounts:
                if self.account == acc.GetUID():
                    GetAliasesAsync(acc, contacts, _success, _error)
                    return
            _error(Exception("Invalid account name"))

        self.teamgeist.ListAccounts(got_accounts, got_error)


    @dbus.service.method(dbus_interface=TEAMGEIST_ACCOUNT_INTERFACE,
                         in_signature='s', out_signature='s',
                         async_callbacks=('_success', '_error'))
    def GetDisplayName(self, account, _success, _error):
        """
        Get the display name of the account
        """
        def got_error(error):
            _error(Exception(error))

        def got_accounts(accounts):
            def got_name(name):
                _success(name)

            for acc in accounts:
                if account == acc.GetUID():
                    acc.GetDisplayName(got_name, got_error)
                    return
            _error(Exception("Invalid account name"))

        self.teamgeist.ListAccounts(got_accounts, got_error)


    @dbus.service.method(dbus_interface=TEAMGEIST_ACCOUNT_INTERFACE,
                         in_signature='s', out_signature='s',
                         async_callbacks=('_success', '_error'))
    def GetNormalizedName(self, account, _success, _error):
        """
        Get the display name of the account
        """
        def got_error(error):
            _error(Exception(error))

        def got_accounts(accounts):
            def got_name(name):
                _success(name)

            for acc in accounts:
                if account == acc.GetUID():
                    acc.GetNormalizedName(got_name, got_error)
                    return
            _error(Exception("Invalid account name"))

        self.teamgeist.ListAccounts(got_accounts, got_error)

    @dbus.service.method(dbus_interface=TEAMGEIST_ACCOUNT_INTERFACE,
                         in_signature='s', out_signature='a(ssu)',
                         async_callbacks=('_success', '_error'))
    def GetContacts(self, account, _success, _error):
        """
        Returns the list of all the contacts available in this account
        Contacts are returned as (uid, alias, presence) structures.
        Contacts are returned as strings and can be used on the
        org.gnome.zeitgeist.Teamgeist.Interfaces.Contact interface
        """
        class GetContactsAsync(object):
            def __init__(self, parent):
                self.parent = parent
                self.parent.teamgeist.ListAccounts(self.got_accounts,
                                                   self.got_error)
            def got_error(self, error):
                _error(Exception(error))

            def got_accounts(self, accounts):
                for acc in accounts:
                    if account == acc.GetUID():
                        acc.GetContacts(self.got_contacts,
                                        self.got_error)
                        return
                _error(Exception("Invalid account name"))
            def got_contacts(self, contacts):
                if contacts:
                    GetContactInfoAsync(contacts, _success, _error)
                else:
                    _error("Account is not connected")
        GetContactsAsync(self)

    @dbus.service.signal(dbus_interface=TEAMGEIST_INTERFACE, signature='ssss')
    def TeamInvitationReceived(self, account, user, type, team_id):
        """
        Signal sent when a user invites you to join a team
        """
        raise NotImplementedError()


if __name__ == "__main__":
    import gobject
    import dbus
    import dbus.glib
    mainloop = gobject.MainLoop(is_running=True)
    dmainloop = dbus.mainloop.glib.DBusGMainLoop()
    dbus.set_default_main_loop(dmainloop)

    TeamgeistService()

    while mainloop.is_running():
        print 'running'
        try:
            mainloop.run()
        except KeyboardInterrupt:
            gobject.idle_add(quit, 1)
        print 'done\n\n'
