# -.- coding: utf-8 -.-

# Teamgeist
#
# Copyright (C) 2009 Collabora Limited
# *  Contact: Youness Alaoui
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# * Contributors:
# *   Youness Alaoui, Collabora Ltd.


import dbus
import dbus.service

TEAMGEIST_INTERFACE = "org.gnome.zeitgeist.Teamgeist"
TEAMGEIST_ACCOUNT_INTERFACE = "org.gnome.zeitgeist.Teamgeist.Interfaces.Account"
TEAMGEIST_CONTACT_INTERFACE = "org.gnome.zeitgeist.Teamgeist.Interfaces.Contact"
TEAM_INTERFACE = "org.gnome.zeitgeist.Teamgeist.Team"
TEAM_MEMBER_INTERFACE = "org.gnome.zeitgeist.Teamgeist.Team.Interfaces.Member"
TEAM_TOPIC_INTERFACE = "org.gnome.zeitgeist.Teamgeist.Team.Interfaces.Topic"

# TODO: Add dbus properties to the interfaces that deserve them

class MemberInterface(dbus.service.Interface):
    """
    This interface allows a user to query information on a member of a team
    """

    @dbus.service.method(dbus_interface=TEAM_MEMBER_INTERFACE,
                         in_signature='as', out_signature='au')
    def GetPresences(self, members):
        """
        Returns the presences of the members
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_MEMBER_INTERFACE,
                         in_signature='as', out_signature='as')
    def GetAliases(self, members):
        """
        Returns the aliases of the members
        """
        # DONE
        raise NotImplementedError()
#TODO add signals for presence/alias changes

class TopicInterface(dbus.service.Interface):
    """
    This interface allows to query a topic for its search criterias
    """

    @dbus.service.method(dbus_interface=TEAM_TOPIC_INTERFACE,
                         in_signature='as', out_signature='aa(suv)')
    def ListSearchCriterias(self, topics):
        """
        Return all the search criterias for each requested topic
        An empty list of search criterias means the topic was invalid.
        """
        # DONE
        raise NotImplementedError()

class TeamInterface(dbus.service.Interface):
    """
    This interface represents a Teamgeist Team
    """

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='', out_signature='s')
    def GetName(self):
        """
        Get the name of the team. This is equivalent to the MUC's name
        """
        # TODO
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='s', out_signature='')
    def SetName(self, name):
        """
        Set the name of the team. This is equivalent to the MUC's name
        """
        # TODO
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='', out_signature='s')
    def GetSubject(self):
        """
        Get the name of the team. This is equivalent to the MUC's subject
        """
        # TODO
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='s', out_signature='')
    def SetSubject(self, subject):
        """
        Set the name of the team. This is equivalent to the MUC's subject
        """
        # TODO
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='', out_signature='a(ssu)')
    def GetMembers(self):
        """
        Returns the list of all the members currently in the team
        Members are returned as (uid, alias, presence) structures.
        The unique identifier string and can be used on the
        org.gnome.zeitgeist.Teamgeist.Team.Interfaces.Member interface
        or for method calls like InviteMember, SuggestTopicToMember, etc..
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='s', out_signature='')
    def InviteMember(self, user):
        """
        Invite a user to become a member of the team.
        If the current user does not have enough priviledges to invite that
        user, then this will send a request to the owner of the team to
        invite that user
        """
        # LATER
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='s', out_signature='')
    def KickMember(self, user):
        """
        Kick a user from the team.
        If the current user does not have enough priviledges to kick that
        user, then this will send a request to the owner of the team to
        kick that user
        """
        # LATER
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='', out_signature='')
    def JoinTeam(self):
        """
        Join the team.
        This should set the team as joinable in the database and attempt to join
        it if the account is available and connected
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='', out_signature='')
    def LeaveTeam(self):
        """
        Leave the team.
        This should leave the MUC and make the team non-joinable in the database
        without deleting it and without deleting the associated topics
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='', out_signature='as')
    def ListTopics(self):
        """
        List all the topics available for the team
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='sa(suv)', out_signature='s')
    def AddTopic(self, topic_name, search_criterias):
        """
        Adds a topic to a team with its associated search criterias
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='s', out_signature='')
    def DeleteTopic(self, topic):
        """
        Deletes a topic from a team
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='ss', out_signature='')
    def SuggestTopicToMember(self, topic, member):
        """
        Suggests to a member of the team a topic for him to share
        """
        # LATER
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='s', out_signature='')
    def SuggestTopicToTeam(self, topic):
        """
        Suggests to all members of the team a topic to be shared
        """
        # LATER
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='ss', out_signature='')
    def RequestDocument(self, member, document):
        """
        Request a document from a member of the team
        """
        # LATER
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature='a{sau}', out_signature='a{sa(asaasay)}')
    def GetEvents(self, event_ids):
        """
        Get the events with the ids {owner: [remote_ids]} requested.
        If the owner is the empty string, then it represents the local user
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature="(xx)a(sasaasay)uu",
                         out_signature="a{sau}")
    def FindEventIds(self, time_range, event_templates, max_events, order):
        """
        Search for events using an event template and return
        the id {owner: [remote_ids]} of the events.
        If the owner is the empty string, then it represents the local user
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAM_INTERFACE,
                         in_signature="", out_signature="u")
    def GetCurrentParticipation(self):
        """ Returns whether or not the user is currently participating
        in the team.
        0: Joining not requested
        1: Participating
        2: Account does not exist
        3: Account is offline
        4: Not a member of the team
        5: Joining
        6: Joining failed
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='s')
    def NameChanged(self, name):
        """
        Signal sent when the name of the team is changed
        """
        # TODO or LATER
        pass

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='s')
    def SubjectChanged(self, subject):
        """
        Signal sent when the subject of the team is changed
        """
        # TODO or LATER
        raise NotImplementedError()

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='s')
    def MemberJoined(self, member):
        """
        Signal sent when a member joins the team
        """
        # DONE
        pass

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='s')
    def MemberLeft(self, member):
        """
        Signal sent when a member has left the team
        """
        # DONE
        pass

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='s')
    def UserRequestsInvitation(self, member):
        """
        Signal sent when a user requests you to invite him to the team
        """
        # LATER
        pass

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='ss')
    def MemberRequestsKick(self, member, to_kick):
        """
        Signal sent when a member requests you to kick someone from
        the team
        """
        # LATER
        pass

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='u')
    def EventShared(self, event_id):
        """
        Signal sent when a local event gets shared with the team
        """
        # DONE
        pass

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='sa(suv)')
    def TopicSuggestionReceived(self, member, search_criterias):
        """
        Signal sent when a team member suggests a topic for you to share.
        Accepting a suggested topic just means calling AddTopic
        """
        # LATER
        pass

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='ss')
    def DocumentRequestReceived(self, member, document):
        """
        Signal sent when a team member requests a document from you
        Accepting a share document request just means calling AddTopic with the
        appropriate search criteria for sharing that document with the team, or
        creating a new 'single' user team with the member to share the document
        with and calling AddTopic on that new team.
        """
        # LATER
        pass

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE, signature='u')
    def ParticipationChanged(self, participating):
        """
        Signal sent when Teamgeist participation in a team changes
        (joining or leaving the team)
        """
        # DONE
        pass

    @dbus.service.signal(dbus_interface=TEAM_INTERFACE,
                         signature='s(asaasay)')
    def EventReceived(self, owner, event):
        """
        Signal sent when an event is received from a team member
        """
        # DONE
        pass

    # TODO: add signals for when the list of authorized members changes?
    # needed to know when someone gets kicked... ?


class ContactInterface(dbus.service.Interface):
    """
    This interface allows a user to query information on a member of a team
    """

    @dbus.service.method(dbus_interface=TEAMGEIST_CONTACT_INTERFACE,
                         in_signature='sas', out_signature='au')
    def GetPresences(self, account, contacts):
        """
        Returns the presences of the contacts
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAMGEIST_CONTACT_INTERFACE,
                         in_signature='sas', out_signature='as')
    def GetAliases(self, account, contacts):
        """
        Returns the aliases of the contacts
        """
        # DONE
        raise NotImplementedError()

class AccountInterface(dbus.service.Interface):
    """
    This interface allows access to an account's information
    """

    @dbus.service.method(dbus_interface=TEAMGEIST_ACCOUNT_INTERFACE,
                         in_signature='s', out_signature='s')
    def GetDisplayName(self, account):
        """
        Get the display name of the account
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAMGEIST_ACCOUNT_INTERFACE,
                         in_signature='s', out_signature='s')
    def GetNormalizedName(self, account):
        """
        Get the normalized name of the account
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAMGEIST_ACCOUNT_INTERFACE,
                         in_signature='s', out_signature='a(ssu)')
    def GetContacts(self, account):
        """
        Returns the list of all the contacts available in this account
        Contacts are returned as (uid, alias, presence) structures.
        Contacts are returned as strings and can be used on the
        org.gnome.zeitgeist.Teamgeist.Interfaces.Contact interface
        """
        # DONE
        raise NotImplementedError()

class TeamgeistInterface(dbus.service.Interface):
    """
    This interface represents the main Teamgeist interface for accessing
    accounts and teams and matching events
    """

    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='', out_signature='as')
    def ListAccounts(self):
        """
        Return the list of all the accounts that are available to be used.
        Accounts are returned as strings and can be used on the
        org.gnome.zeitgeist.Teamgeist.Interfaces.Account interface
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='', out_signature='a(ss)')
    def ListTeamTypes(self):
        """
        List all the teams types that are available.
        The team type is needed for adding new teams
        Returns an array of (team_type, description)
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='', out_signature='ao')
    def ListTeams(self):
        """
        List all the teams that the user has.
        Teams are returned as objects implementing the
        org.gnome.zeitgeist.Teamgeist.Team interface
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='sss', out_signature='o')
    def AddTeam(self, account, type, uid):
        """
        This adds the team to the Teams database or makes it joinable if already
        there.
        A org.gnome.zeitgeist.Teamgeist.NewTeam signal will be sent when the team
        is added.
        Returns the object path of the Team
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='o', out_signature='')
    def PurgeTeam(self, team):
        """
        This purge a team from the Teams database deleting all its associated
        topics and events
        A org.gnome.zeitgeist.Teamgeist.TeamPurged signal will be sent when
        the team is deleted.
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='(asaasay)', out_signature='ao')

    def MatchEvent(self, event):
        """
        Match an event with all topics and share that event with the
        appropriate teams if the event matches topics
        Returns the list of teams with which this event got shared
        """
        # DONE
        raise NotImplementedError()

    @dbus.service.signal(dbus_interface=TEAMGEIST_INTERFACE, signature='ssss')
    def TeamInvitationReceived(self, account, user, type, team_id):
        """
        Signal sent when a user invites you to join a team
        """
        # LATER
        pass

    @dbus.service.signal(dbus_interface=TEAMGEIST_INTERFACE, signature='o')
    def NewTeam(self, team):
        """
        Signal sent when a new team is added/created
        """
        # DONE
        pass

    @dbus.service.signal(dbus_interface=TEAMGEIST_INTERFACE, signature='o')
    def TeamPurged(self, team):
        """
        Signal sent when a team has been purged
        """
        # DONE
        pass
