
from signals import SignalDispatcher

from tp_classes import AccountManager
from tp_client import TeamgeistClient
from utils import do_reply
from bases import ReadyableObject
from team import Team

# NOTE: we assume only one teamgeist instance will ever exist/modify the database
# this is necessary in order to add a caching mechanism to avoid doing lots of
# dbus calls and database queries for every call made to the engine

class Contacts(object):
    """
    This class represents a list of contacts
    """
    def __init__(self, connection, handles):
        self.connection = connection
        self.handles = handles

    def GetUIDs(self, reply_handler, error_handler):
        """
        Returns the unique identifier of the contacts
        """
        def got_reply(uids):
            do_reply(reply_handler, uids)

        self.connection.inspect_handles(self.handles,
                                        got_reply, error_handler)

    def GetPresences(self, reply_handler, error_handler):
        """
        Returns the presences of the contacts
        """
        def got_reply(presences):
            result = []
            for handle in self.handles:
                result.append(presences[handle][0])
            do_reply(reply_handler, result)

        self.connection.get_presences(self.handles,
                                        got_reply, error_handler)

    def GetAliases(self, reply_handler, error_handler):
        """
        Returns the aliases of the contacts
        """
        def got_reply(aliases):
            result = []
            for handle in self.handles:
                result.append(aliases[handle])
            do_reply(reply_handler, result)

        self.connection.get_aliases(self.handles,
                                        got_reply, error_handler)

class Member(object):
    """
    This class represents a member of a team
    """
    def __init__(self, connection, handle):
        self.contacts = Contacts(connection, [handle])

    def GetUID(self, reply_handler, error_handler):
        """
        Returns the unique identifier of the member
        """
        def got_reply(uids):
            do_reply(reply_handler, uids[0])

        self.contacts.GetUIDs(got_reply, error_handler)

    def GetPresence(self, reply_handler, error_handler):
        """
        Returns the presence of the member
        """
        def got_reply(presences):
            do_reply(reply_handler, presences[0])

        self.contacts.GetPresences(got_reply, error_handler)

    def GetAlias(self, reply_handler, error_handler):
        """
        Returns the alias of the member
        """
        def got_reply(aliases):
            do_reply(reply_handler, aliases[0])

        self.contacts.GetAliases(got_reply, error_handler)

class Account(object):
    """
    This class represents an account of the user with which he can join a team
    """
    def __init__(self, account_uid, account_obj):
        self._uid = account_uid
        self._obj = account_obj

    def GetUID(self):
        """
        Return a unique identifier for the account
        """
        return self._uid

    def GetNormalizedName(self, reply_handler, error_handler):
        """
        Get the normalized name of the account
        """
        def get_normalized_name():
            do_reply(reply_handler, self._obj.normalized_name)

        if self._obj:
            self._obj.call_when_ready(get_normalized_name)
        else:
            do_reply(error_handler, Exception("Account not available"))



    def GetDisplayName(self, reply_handler, error_handler):
        """
        Get the display name of the account
        """
        def get_display_name():
            do_reply(reply_handler, self._obj.display_name)

        if self._obj:
            self._obj.call_when_ready(get_display_name)
        else:
            do_reply(error_handler, Exception("Account not available"))

    def GetContacts(self, reply_handler, error_handler):
        """
        Returns the list of all the contacts available in this account
        """
        def get_contacts():
            def get_stored_members_cb(handles):
                do_reply(reply_handler, Contacts(self._obj.connection, handles))

            if self._obj:
                if self._obj.connection is None:
                    do_reply(reply_handler, None)
                else:
                    self._obj.connection.get_stored_members(
                        get_stored_members_cb,
                        error_handler)
            else:
                do_reply(error_handler, Exception("Account not available"))

        if self._obj:
            self._obj.call_when_ready(get_contacts)
        else:
            do_reply(error_handler, Exception("Account not available"))


    def GetConnection(self, reply_handler, error_handler):
        def get_connection():
            do_reply(reply_handler, self._obj.connection)

        if self._obj:
            self._obj.call_when_ready(get_connection)
        else:
            do_reply(error_handler, Exception("Account not available"))

class EventSubject(object):
    def __init__(self, event_id, id, interpretation, manifestation,
                 origin, mimetype, text, storage):
        self.id = self.str(id)
        self.event_id = event_id
        self.interpretation = self.str(interpretation)
        self.manifestation = self.str(manifestation)
        self.origin = self.str(origin)
        self.mimetype = self.str(mimetype)
        self.text = self.str(text)
        self.storage = self.str(storage)

    def str(self, obj):
        return unicode(obj) if obj else u""

    @property
    def struct(self):
        return (self.id,
                self.interpretation,
                self.manifestation,
                self.origin,
                self.mimetype,
                self.text,
                self.storage)

    @staticmethod
    def CreateFromStruct(event_id, subject):
        return EventSubject(event_id, subject[0],
                            subject[1], subject[2],
                            subject[3], subject[4],
                            subject[5], subject[6])

class RemoteEvent(object):
    def __init__(self, team, owner, added_timestamp, remote_id,
                 timestamp, interpretation, manifestation,
                 actor, payload, subjects):
        self.id = int(remote_id)
        self.team = team
        self.owner = owner
        self.added = added_timestamp and long(added_timestamp)
        self.timestamp = timestamp and long(timestamp)
        self.interpretation = self.str(interpretation)
        self.manifestation = self.str(manifestation)
        self.actor = self.str(actor)
        self.payload = payload or u""
        self._subjects = subjects

    def str(self, obj):
        return unicode(obj) if obj else u""

    def __repr__(self):
        return str(self.struct)

    @property
    def subjects(self):
        if len(self._subjects) > 0:
            return self._subjects
        else:
            # TODO : a bit hacky.. but in theory an event hsould have at least
            # one subject...
            return [EventSubject(self.id, 0, None, None, None,
                                 None, None, None)]

    @property
    def struct(self):
        return ((self.str(self.id),
                 self.str(self.timestamp),
                 self.interpretation,
                 self.manifestation,
                 self.actor),
                [subject.struct for subject in self.subjects],
                self.payload)

    @staticmethod
    def CreateFromStruct(team, owner, struct):
        event = struct[0]
        payload = struct[2]
        subjects = []
        for subject in struct[1]:
            subjects.append(EventSubject.CreateFromStruct(event[0], subject))
        return RemoteEvent(team, owner, None, event[0],
                           event[1], event[2],event[3],
                           event[4], payload, subjects)
