# -.- coding: utf-8 -.-

# Teamgeist
#
# Copyright (C) 2009 Collabora Limited
# *  Contact: Youness Alaoui
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# * Contributors:
# *   Youness Alaoui, Collabora Ltd.


from signals import SignalDispatcher
from utils import do_reply
from bases import ReadyableObject
import string
import sqlite3
import copy

MATCH_EQUALS = 0
MATCH_NOT_EQUALS = 1
MATCH_WILDCARD_SEARCH = 2
    
class SearchCriteria(object):
    """
    This class represents a search criteria.
    A search criteria can be used to define any type of search or sharable
    element for teamgeist. The 'criteria' attribute determines what kind of
    search criteria it is, while the 'match' attribute determines how the
    criteria should be matched to the value.
    """

    # A search string for a tag from tracker
    TRACKER_TAG = 'tracker_tag'
    # A document path to be shared with the team
    SHARED_DOCUMENT = 'shared_document'
    # An Event URI to be shared with the team
    EVENT = 'event'
    # An Application name for which events must be shared with the team
    APPLICATION = 'application'
    # The path of a file for which events must be shared with the team
    PATH = 'path'
    # Annotation to be present on an event to be shared with the team
    ANNOTATION = 'annotation'
    # TODO: add more search criterias depending on what needs to be searched for


    def __init__(self, id, criteria, match, value):
        self.id = id
        self.criteria = criteria
        self.match = match
        self.value = value
        self.topic = None

    def __eq__(self, obj):
        return self.id == obj.id

    def __repr__(self):
        return repr((self.id, self.criteria, self.match, self.value))

    def MatchEvent(self, event):
        """
        Match an event with the search criteria.
        Return True if the event matches the criteria, False otherwise
        """
        value = None
        # TODO: use criteria constants
        if self.criteria == 'event_id':
            value = event.id
        elif self.criteria == 'timerange':
            try:
                start = self.value.split(":")[0]
                end = self.value.split(":")[1]
                if event.timestamp > start and event.timestamp < end:
                    if self.match is MATCH_EQUALS:
                        return True
                    else:
                        return False
            except:
                return False
        elif self.criteria in ('interpretation',
                        'manifestation',
                        'actor',
                        'payload'):
            value = getattr(event, self.criteria, None)
        elif self.criteria.startswith("subject_"):
            for subject in event.subjects:
                if self.MatchSubject(subject):
                    return True
            return False
        if value:
            if self.match is MATCH_EQUALS:
                return self.value == value
            elif self.match is MATCH_NOT_EQUALS:
                return self.value != value
            else:
                bool = string.find(value, self.value)
                if bool > -1:
                    bool = True
                else: 
                    bool = string.find(self.value, value)
                    if bool > 0:
                        bool = True
                    else: 
                        bool = False

                return bool
        else:
            return False

    def MatchSubject(self, subject):
        """
        Match an event's subject with the search criteria.
        Return True if the subject matches the criteria, False otherwise
        """

        if not self.criteria.startswith("subject_"):
            return False
        criteria = self.criteria[len("subject_"):]

        value = None
        if criteria in ('id',
                        'interpretation',
                        'manifestation',
                        'origin',
                        'mimetype',
                        'text',
                        'storage'):
            value = getattr(subject, criteria, None)

        if value:
            if self.match is MATCH_EQUALS:
                return self.value == value
            elif self.match is MATCH_NOT_EQUALS:
                return self.value != value
            else:
                print "unable to match with match value %d" % self.match
                return False
        else:
            return False

class Topic(object):
    """
    This class represents a Teamgeist Topic and contains the search criterias
    for this topic.
    """
    def __init__(self, id, name, search_criterias, team):
        self.id = id
        self.name = name
        self.team = team
        self.search_criterias = search_criterias
        for search in self.search_criterias:
            search.topic = self

    def __eq__(self, obj):
        return self.id == obj.id

    def __repr__(self):
        return "<Topic %d:%s (team %d): %r>"  % (self.id, self.name,
                                                 self.team.id,
                                                 self.search_criterias)

    def ListSearchCriterias(self):
        """ Return all the search criterias for this topic """
        return self.search_criterias

    def GetTeam(self):
        """ Return the team this topic is associated with """
        return self.team

    def MatchEvent(self, event):
        """
        Match an event with the topic.
        Return True if the topic matches the criteria, False otherwise.
        For an event to match a topic, it must match all search criterias defined
        in the topic
        """
        matches = True

        # If it's an empty topic, then it doesn't match
        if len(self.search_criterias) == 0:
            matches = False

        # match only if all search criterias match
        for criteria in self.search_criterias:
            matches = matches and criteria.MatchEvent(event)
            if matches is False:
                break
        return matches

class Team(SignalDispatcher, ReadyableObject):
    """
    This class represents a Teamgeist Team
    """
    PARTICIPATION_NOT_REQUESTED = 0
    PARTICIPATION_JOINED = 1
    PARTICIPATION_ACCOUNT_NOT_EXIST = 2
    PARTICIPATION_ACCOUNT_OFFLINE = 2
    PARTICIPATION_NOT_A_MEMBER = 3
    PARTICIPATION_JOINING = 4
    PARTICIPATION_FAILED = 5

    def __init__(self, teamgeist, id, account, type, uid, join):
        """
        Initialize the instance by specifying the type of team (single XMPP user,
        or XMPP muc, or ...) and the unique identifier for the Team.
        Also specify whether Teamgeist is currently participating in the team.
        This could be either because the team is not joinable (you left, or got
        kicked), or because no account is connected.
        """
        print "--------"
        print teamgeist, id, account, type, uid, join
        print "--------"
        
        SignalDispatcher.__init__(self)
        ReadyableObject.__init__(self)

        self.id = id
        self.account = account
        self.type = type
        self.uid = uid
        self.join = join
        self.teamgeist = teamgeist

        self._topics = []
        self._populate_topics()

        if type in self._type_handlers:
            self._handler = self._type_handlers[type](self)
        else:
            raise Exception("Invalid team type")

        def event_received(emitter, owner, event):
            self.teamgeist._dbaccessor.InsertEvent(self, event)
            self.emit_signal("EventReceived", owner, event)
        self._handler.connect("EventReceived", event_received)

        self._handler.connect("ParticipationChanged",
                              lambda emitter, participation:
                                  self.emit_signal("ParticipationChanged",
                                                   participation))

        self._handler.connect("MemberJoined",
                              lambda emitter, member:
                                  self.emit_signal("MemberJoined", member))

        self._handler.connect("MemberLeft",
                              lambda emitter, member:
                                  self.emit_signal("MemberLeft", member))

    def __eq__(self, obj):
        if obj == None:
            return False
        elif isinstance(obj, int):
            return self.id == obj
        else:
            return self.id == obj.id

    def _populate_topics(self):
        topics = self.teamgeist._dbaccessor.GetTopics(self)

        # Add newly created teams to our list
        for topic in topics:
            topic = self._new_topic(topic[0], topic[1], topic[2])
            self._topics.append(topic)


        self.ready()

    def _new_topic(self, topic_id, name, searchs):
        search_criterias = []
        for search in searchs:
            if len(search) == 3:
                search_criteria = SearchCriteria(0, search[0],
                                                 search[1], search[2])
            elif len(search) == 4:
                search_criteria = SearchCriteria(search[0], search[1],
                                                 search[2], search[3])
            else:
                raise Exception("Invalid search criteria length")
            search_criterias.append(search_criteria)
        return Topic(topic_id, name, search_criterias, self)

    def GetName(self, reply_handler, error_handler):
        """
        Get the name of the team. This is equivalent to the MUC's name
        """
        return self._handler.GetName(reply_handler, error_handler)

    def SetName(self, name, reply_handler, error_handler):
        """
        Set the name of the team. This is equivalent to the MUC's name
        """
        return self._handler.SetName(name, reply_handler, error_handler)

    def GetSubject(self, reply_handler, error_handler):
        """
        Get the name of the team. This is equivalent to the MUC's subject
        """
        return self._handler.GetSubject(reply_handler, error_handler)

    def SetSubject(self, subject, reply_handler, error_handler):
        """
        Set the name of the team. This is equivalent to the MUC's subject
        """
        return self._handler.SetSubject(subject, reply_handler, error_handler)

    def GetMembers(self, reply_handler, error_handler):
        """
        Returns the list of all the members currently in the team
        """
        return self._handler.GetMembers(reply_handler, error_handler)

    def InviteMember(self, user, reply_handler, error_handler):
        """
        Invite a user to become a member of the team.
        If the current user does not have enough priviledges to invite that
        user, then this will send a request to the owner of the team to
        invite that user
        """
        return self._handler.InviteMember(user)

    def KickMember(self, user, reply_handler, error_handler):
        """
        Kick a user from the team.
        If the current user does not have enough priviledges to kick that
        user, then this will send a request to the owner of the team to
        kick that user
        """
        return self._handler.KickMember(user, reply_handler, error_handler)

    def JoinTeam(self, reply_handler, error_handler):
        """
        Join the team.
        This should set the team as joinable in the database and attempt to join
        it if the account is available and connected
        """
        if self.join is False:
            self.teamgeist._dbaccessor.SetTeamJoinable(self, True)
            self.join = True
        # TODO: listen to the account's ConnectionChanged and retry to join
        # if we get connected.
        return self._handler.JoinTeam(reply_handler, error_handler)

    def LeaveTeam(self, reply_handler, error_handler):
        """
        Leave the team.
        This should leave the MUC and make the team non-joinable in the database
        without deleting it and without deleting the associated topics
        """
        if self.join is True:
            self.teamgeist._dbaccessor.SetTeamJoinable(self, False)
            self.join = False
        return self._handler.LeaveTeam(reply_handler, error_handler)

    def ListTopics(self, reply_handler, error_handler):
        """
        List all the topics available for the team
        """
        def list_topics_ready():
            do_reply(reply_handler, self._topics)

        self.call_when_ready(list_topics_ready)


    def AddTopic(self, name, search_criterias, reply_handler, error_handler):
        """
        Adds a topic to a team with its associated search criterias
        """
        def add_topic_ready():
            topic = self._new_topic(0, name, search_criterias)
            try:
                id = self.teamgeist._dbaccessor.AddTopic(self, topic)
            except sqlite3.IntegrityError:
                do_reply(error_handler,
                         Exception("Topic with similar name already exists"))
                return
            if id == 0:
                # TODO: figure out why (empty searches?)
                do_reply(error_handler,
                         Exception("Could not add topic to database"))
            else:
                self._topics.append(topic)
                do_reply(reply_handler, topic)

        self.call_when_ready(add_topic_ready)


    def DeleteTopic(self, topic, reply_handler, error_handler):
        """
        Deletes a topic from a team
        """
        def delete_topic_ready():
            if topic in self._topics:
                self.teamgeist._dbaccessor.DeleteTopic(topic)
                self._topics.remove(topic)
                do_reply(reply_handler)
            else:
                do_reply(error_handler,
                         Exception("Topic does not belong to this team"))

        self.call_when_ready(delete_topic_ready)

    def SuggestTopicToMember(self, topic, member, reply_handler, error_handler):
        """
        Suggests to a member of the team a topic for him to share
        """
        return self._handler.SuggestTopicToMember(topic, member,
                                                  reply_handler, error_handler)

    def SuggestTopicToTeam(self, topic, reply_handler, error_handler):
        """
        Suggests to all members of the team a topic to be shared
        """
        return self._handler.SuggestTopicToTeam(topic,
                                                reply_handler, error_handler)

    def RequestDocument(self, member, document, reply_handler, error_handler):
        """
        Request a document from a member of the team
        """
        return self._handler.RequestDocument(member, document,
                                             reply_handler, error_handler)

    def MatchEvent(self, event, reply_handler, error_handler):
        """
        Match an event with the team's topics.
        Return True if the topic matches the team's topics, False otherwise.
        For an event to match, it must match at least one topic.
        """
        def got_ready():
            for topic in self._topics:
                if topic.MatchEvent(event):
                    new_event = copy.copy(event)
                    new_event.team = self
                    self.teamgeist._dbaccessor.InsertEvent(self, new_event)
                    self._handler.ShareEvent(new_event)
                    self.emit_signal("EventShared", new_event)
                    do_reply(reply_handler, True)
                    return
            do_reply(reply_handler, True)

        self.call_when_ready(got_ready)

    def GetEvents(self, event_ids):
        """
        Get the events with the ids requested.
        """
        return dict([(owner,
                      self.teamgeist._dbaccessor.GetEvents(self,
                                                           owner,
                                                           event_ids[owner]))
                     for owner in event_ids.keys()])

    def FindEventIds(self, time_range, event_templates, max_events, order):
        """
        Search for events using an event template
        """
        return self.teamgeist._dbaccessor.FindEventIds(self, time_range,
                                                       event_templates,
                                                       max_events, order)

    def GetCurrentParticipation(self):
        """ Returns whether or not the user is currently participating
        in the team
        """
        return self._handler.GetCurrentParticipation()

    _type_handlers = {}
    @staticmethod
    def ListTypes():
        return [(type, Team._type_handlers[type].GetDescription())
                for type in Team._type_handlers.keys()]

    @staticmethod
    def RegisterType(type, cls):
        Team._type_handlers[type] = cls

    @staticmethod
    def ValidateType(type):
        return type in Team._type_handlers

    __signals__ = {
        "NameChanged": # TODO
            "Signal sent when the name of the team is changed",
        "SubjectChanged": # TODO
            "Signal sent when the subject of the team is changed",
        "MemberJoined": # TODO
            "Signal sent when a member joins the team",
        "MemberLeft": # TODO
            "Signal sent when a member has left the team",
        "UserRequestsInvitation": # LATER
            "Signal sent when a user requests you to invite him to the team",
        "MemberRequestsKick": # LATER
            """Signal sent when a member requests you to kick someone from
            the team
            """,
        "EventShared":
            "Signal sent when a local event gets shared with the team",
        "TopicSuggestionReceived": # LATER
            "Signal sent when a team member suggests a topic for you to share",
        "DocumentRequestReceived": # LATER
            "Signal sent when a team member requests a document from you",
        "ParticipationChanged":
            """Signal sent when Teamgeist participation in a team changes
            (joining or leaving the team)
            """,
        "EventReceived":
            "Signal sent when an event is received from a team member",
        }


    # TODO: add signals for when the list of authorized members changes?
    # needed to know when someone gets kicked... ?


# We need to import this after we defined the Team class since it's needed
# by the team_type
import xmpp_tube
