# -.- coding: utf-8 -.-

# Teamgeist
#
# Copyright (C) 2009 Collabora Limited
# *  Contact: Youness Alaoui
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# * Contributors:
# *   Youness Alaoui, Collabora Ltd.


from signals import SignalDispatcher

from tp_classes import AccountManager
from tp_client import TeamgeistClient
from utils import do_reply
from bases import ReadyableObject
from team import Team

# NOTE: we assume only one teamgeist instance will ever exist/modify the database
# this is necessary in order to add a caching mechanism to avoid doing lots of
# dbus calls and database queries for every call made to the engine

class Contacts(object):
    """
    This class represents a list of contacts
    """
    def __init__(self, connection, handles):
        self.connection = connection
        self.handles = handles

    def GetUIDs(self, reply_handler, error_handler):
        """
        Returns the unique identifier of the contacts
        """
        def got_reply(uids):
            do_reply(reply_handler, uids)

        self.connection.inspect_handles(self.handles,
                                        got_reply, error_handler)

    def GetPresences(self, reply_handler, error_handler):
        """
        Returns the presences of the contacts
        """
        def got_reply(presences):
            result = []
            for handle in self.handles:
                result.append(presences[handle][0])
            do_reply(reply_handler, result)

        self.connection.get_presences(self.handles,
                                        got_reply, error_handler)

    def GetAliases(self, reply_handler, error_handler):
        """
        Returns the aliases of the contacts
        """
        def got_reply(aliases):
            result = []
            for handle in self.handles:
                result.append(aliases[handle])
            do_reply(reply_handler, result)

        self.connection.get_aliases(self.handles,
                                        got_reply, error_handler)

class Member(object):
    """
    This class represents a member of a team
    """
    def __init__(self, connection, handle):
        self.contacts = Contacts(connection, [handle])

    def GetUID(self, reply_handler, error_handler):
        """
        Returns the unique identifier of the member
        """
        def got_reply(uids):
            do_reply(reply_handler, uids[0])

        self.contacts.GetUIDs(got_reply, error_handler)

    def GetPresence(self, reply_handler, error_handler):
        """
        Returns the presence of the member
        """
        def got_reply(presences):
            do_reply(reply_handler, presences[0])

        self.contacts.GetPresences(got_reply, error_handler)

    def GetAlias(self, reply_handler, error_handler):
        """
        Returns the alias of the member
        """
        def got_reply(aliases):
            do_reply(reply_handler, aliases[0])

        self.contacts.GetAliases(got_reply, error_handler)

class Account(object):
    """
    This class represents an account of the user with which he can join a team
    """
    def __init__(self, account_uid, account_obj):
        self._uid = account_uid
        self._obj = account_obj

    def GetUID(self):
        """
        Return a unique identifier for the account
        """
        return self._uid

    def GetNormalizedName(self, reply_handler, error_handler):
        """
        Get the normalized name of the account
        """
        def get_normalized_name():
            do_reply(reply_handler, self._obj.normalized_name)

        if self._obj:
            self._obj.call_when_ready(get_normalized_name)
        else:
            do_reply(error_handler, Exception("Account not available"))



    def GetDisplayName(self, reply_handler, error_handler):
        """
        Get the display name of the account
        """
        def get_display_name():
            do_reply(reply_handler, self._obj.display_name)

        if self._obj:
            self._obj.call_when_ready(get_display_name)
        else:
            do_reply(error_handler, Exception("Account not available"))

    def GetContacts(self, reply_handler, error_handler):
        """
        Returns the list of all the contacts available in this account
        """
        def get_contacts():
            def get_stored_members_cb(handles):
                do_reply(reply_handler, Contacts(self._obj.connection, handles))

            if self._obj:
                if self._obj.connection is None:
                    do_reply(reply_handler, None)
                else:
                    self._obj.connection.get_stored_members(
                        get_stored_members_cb,
                        error_handler)
            else:
                do_reply(error_handler, Exception("Account not available"))

        if self._obj:
            self._obj.call_when_ready(get_contacts)
        else:
            do_reply(error_handler, Exception("Account not available"))


    def GetConnection(self, reply_handler, error_handler):
        def get_connection():
            do_reply(reply_handler, self._obj.connection)

        if self._obj:
            self._obj.call_when_ready(get_connection)
        else:
            do_reply(error_handler, Exception("Account not available"))



class EventSubject(object):
    def __init__(self, event_id, id, interpretation, manifestation,
                 origin, mimetype, text, storage):
        self.id = self.str(id)
        self.event_id = event_id
        self.interpretation = self.str(interpretation)
        self.manifestation = self.str(manifestation)
        self.origin = self.str(origin)
        self.mimetype = self.str(mimetype)
        self.text = self.str(text)
        self.storage = self.str(storage)

    def str(self, obj):
        return unicode(obj) if obj else u""

    @property
    def struct(self):
        return (self.id,
                self.interpretation,
                self.manifestation,
                self.origin,
                self.mimetype,
                self.text,
                self.storage)

    @staticmethod
    def CreateFromStruct(event_id, subject):
        return EventSubject(event_id, subject[0],
                            subject[1], subject[2],
                            subject[3], subject[4],
                            subject[5], subject[6])

class RemoteEvent(object):
    def __init__(self, team, owner, added_timestamp, remote_id,
                 timestamp, interpretation, manifestation,
                 actor, payload, subjects):
        self.id = int(remote_id)
        self.team = team
        self.owner = owner
        self.added = added_timestamp and long(added_timestamp)
        self.timestamp = timestamp and long(timestamp)
        self.interpretation = self.str(interpretation)
        self.manifestation = self.str(manifestation)
        self.actor = self.str(actor)
        self.payload = payload or u""
        self._subjects = subjects

    def str(self, obj):
        return unicode(obj) if obj else u""

    def __repr__(self):
        return str(self.struct)

    @property
    def subjects(self):
        if len(self._subjects) > 0:
            return self._subjects
        else:
            # TODO : a bit hacky.. but in theory an event hsould have at least
            # one subject...
            return [EventSubject(self.id, 0, None, None, None,
                                 None, None, None)]

    @property
    def struct(self):
        return ((self.str(self.id),
                 self.str(self.timestamp),
                 self.interpretation,
                 self.manifestation,
                 self.actor),
                [subject for subject in self.subjects],
                self.payload)

    @staticmethod
    def CreateFromStruct(team, owner, struct):
        event = struct[0]
        payload = struct[2]
        subjects = []
        for subject in struct[1]:
            subjects.append(EventSubject.CreateFromStruct(event[0], subject))
        return RemoteEvent(team, owner, None, event[0],
                           event[1], event[2],event[3],
                           event[4], payload, subjects)

from dbaccessor import DBAccessor
class Teamgeist(SignalDispatcher, ReadyableObject):
    """
    This class represents the public API of Teamgeist.
    It allows access to all of Teamgeist's functionality allowing you to
    create/join/name teams, create/delete topics, match/share events
    """
    def __init__(self):
        SignalDispatcher.__init__(self)
        ReadyableObject.__init__(self)
        self._teams = []
        self._dbaccessor = DBAccessor()
        self._accountmanager = AccountManager()
        self._tp_client = TeamgeistClient()
        self._accounts = {}
        self._populate_accounts()

        # TODO: listen to TeamgeistClient for incoming team invitation
        # and send the signal TeamInvitationReceived

    def _populate_accounts(self):
        self._accountmanager.call_when_ready(self._populate_accounts_ready)

    def _populate_accounts_ready(self):
        accounts = self._accountmanager.accounts

        # Remove deleted accounts from our list
        for account in self._accounts.keys():
            if account not in accounts.keys():
                del self._accounts[account]

        # Add newly created accounts to our list
        for account in accounts.keys():
            if account not in self._accounts.keys():
                self._accounts[str(account)] = Account(str(account),
                                                       accounts[account])
            else:
                self._accounts[str(account)]._obj = accounts[account]
        self._accountmanager.connect("NewAccount", self._new_account)
        self._accountmanager.connect("AccountDeleted", self._account_deleted)

        self._populate_teams()

    def _new_account(self, emitter, account, obj):
        self._accounts[str(account)] = Account(str(account), obj)

    def _account_deleted(self, emitter, account, obj):
        del self._accounts[account]

    def _populate_teams(self):
        teams = self._dbaccessor.GetTeams()

        for team in teams:
            try:
                self._new_team(team[0], team[1], team[2], team[3], team[4])
            except Exception, e:
                # Invalid team? maybe the database has a wrong team type
                print "Unable to instantiate Team : ", e

        self.ready()

    def _new_team(self, id, account_id, type, uid, joinable):
        account = None
        for acc in self._accounts.values():
            if acc.GetUID() == account_id:
                account = acc
                break
        if account is None:
            account = Account(account_id, None)
            self._accounts[account_id] = account
        new_team = Team(self, id, account, type, uid, joinable)
        self._teams.append(new_team)
        self.emit_signal("NewTeam", new_team)
        if joinable:
            new_team.JoinTeam(None, None)
        return new_team


    def ListAccounts(self, reply_handler, error_handler):
        """
        Return the list of all the accounts that are available to be used
        """
        def list_accounts_ready():
            do_reply(reply_handler, self._accounts.values())

        self.call_when_ready(list_accounts_ready)

    def ListTeamTypes(self):
        """
        List all the teams types that are available.
        The team type is needed for adding new teams
        Returns an array of (team_type, description)
        """
        return Team.ListTypes()

    def ListTeams(self, reply_handler, error_handler):
        """
        List all the teams that the user has
        """
        # TODO: if we modify this list, others have access to it..
        # so if we do ListTeams-> for team in teams: PurgeTeam, it won't work
        # because the 'teams' in the for loop will be modified, and we won't purge
        # all the teams!!!!
        def list_teams_ready():
            do_reply(reply_handler, self._teams)

        self.call_when_ready(list_teams_ready)

    def AddTeam(self, account, type, uid, reply_handler, error_handler):
        """
        This adds the team to the Teams database or makes it joinable if already
        there.
        """
        def add_team_ready():
            if Team.ValidateType(type) is False:
                do_reply(error_handler, Exception("Invalid team type"))
                return
            id = self._dbaccessor.AddTeam(account, type, uid, True)
            if id == 0:
                # There's a use case where the account type is invalid
                # the team doesn't exist in teamgeist since it can't be
                # instanciated, but it exists in the database so we get a
                # "Team already exists" error although the team cannot be found
                do_reply(error_handler, Exception("Team already exists"))
            else:
                try:
                    team = self._new_team(id, account, type, uid, True)
                    do_reply(reply_handler, team)
                except Exception, e:
                    do_reply(error_handler, e)

        self.call_when_ready(add_team_ready)


    def PurgeTeam(self, team, reply_handler, error_handler):
        """
        This purge a team from the Teams database deleting all its associated
        topics and events
        """
        def purge_team_ready():
            print "ready to purge team ", team
            if team.id == 0:
                do_reply(error_handler, Exception("Team does not exist"))
            else:
                self._dbaccessor.DeleteTeam(team)
                self.emit_signal("TeamPurged", team)
                team.id = 0
                team.join = False
                team.LeaveTeam(None, None)
                self._teams.remove(team)
                do_reply(reply_handler)

        print "Call when reading : purge team ", team
        self.call_when_ready(purge_team_ready)
        print "done : purge team ", team


    def ListTopics(self, reply_handler, error_handler):
        """
        List all the topics available in every team
        """
        def list_topics_ready():
            topics_dict = {}
            def team_topics_reply(num_teams, team, topics):
                topics_dict[team] = topics
                if len(topics_dict) == num_teams:
                    result = []
                    for topics in topics_dict.values():
                        result.extend(topics)
                    do_reply(reply_handler, result)

            def team_topics_error(num_teams, team, error):
                # An error should never happen here anyways, so let's not care
                # about the error variable, and let's not care about calling
                # error_handler if an element in topics_dict is None.
                # FIXME: maybe still make it cleaner?
                topics_dict[team] = None
                if len(topics_dict) == num_teams:
                    result = []
                    for topics in topics_dict.values():
                        result.extend(topics)
                    do_reply(reply_handler, result)

            for team in self._teams:
                team.ListTopics((team_topics_reply, len(self._teams), team),
                                (team_topics_error, len(self._teams), team))

            if len(self._teams) == 0:
                do_reply(reply_handler, [])

        self.call_when_ready(list_topics_ready)

    def MatchEvent(self, event):
        """
        Match an event with all topics and share that event with the
        appropriate teams if the event matches topics
        """
        def match_event_ready():
            # TODO: return the list of matched teams
            for team in self._teams:
                team.MatchEvent(event, None, None)

        self.call_when_ready(match_event_ready)

    __signals__ = {
        "TeamInvitationReceived": # NotImplemented
            "Signal sent when a user invites you to join a team",
        "NewTeam":
            "Signal sent when a new team is added/created",
        "TeamPurged":
            "Signal sent when a team has been purged",
        }


if __name__ == "__main__":
    def testAccounts():
        def reply(accounts):
            def reply2(name):
                print name
            def reply3(contacts):
                def reply4(uids):
                    print map(str, uids)
                def reply5(aliases):
                    print map(unicode, aliases)
                def reply6(presences):
                    print map(int, presences)
                if contacts is not None:
                    contacts.GetUIDs(reply4, None)
                    contacts.GetAliases(reply5, None)
                    contacts.GetPresences(reply6, None)
            for account in accounts:
                account.GetDisplayName(reply2, None)
                account.GetNormalizedName(reply2, None)
                account.GetContacts(reply3, None)

        t.ListAccounts(reply, None)

    def testAccounts2():
        def idler():
            def reply(accounts):
                print accounts

            t.ListAccounts(reply, None)
            return True
        gobject.timeout_add(5000, idler)

    def testTeams():
        def reply(teams):
            print "Teams are : ", teams
        def reply2(team):
            print "New team is ", team
            t.PurgeTeam(team, (reply4, team), error2)
            t.PurgeTeam(team, (reply4, team), error2)
            t.ListTeams(reply, None)
        def reply3(team):
            print "New team is ", team
        def reply4(team):
            print "Purged team is ", team
        def error(err):
            print "unable to add team : ", err
        def error2(err):
            print "unable to purge team : ", err

        print "Team types are : ", t.ListTeamTypes()

        t.ListTeams(reply, None)
        t.AddTeam("youness.alaoui@collabora.co.uk", "xmpp_muc",
                         "test@conference.jabber.org", reply3, error)
        t.ListTeams(reply, None)
        t.AddTeam("kakaroto@collabora.co.uk", "xmpp_muc",
                  "test@conference.jabber.com", reply3, error)
        t.AddTeam("kakaroto@collabora.co.uk", "xmpp_muc",
                  "test@conference.jabber.com", reply3, error)
        t.AddTeam("youness.alaoui@collabora.co.uk", "xmpp_muc",
                         "test@conference.jabber.com", reply2, error)
        t.AddTeam("/org/freedesktop/Telepathy/Account/gabble/jabber/kakaroto_2etest1_40test_2ecollabora_2eco_2euk0", "xmpp_muc",
                         "teamgeist@conference.jabber.com", reply3, error)
        t.AddTeam("/org/freedesktop/Telepathy/Account/gabble/jabber/kakaroto_2etest1_40test_2ecollabora_2eco_2euk0", "xmpp_muc",
                         "teamgeist@conference.jabber.com", reply3, error)

        t.ListTeams(reply, None)

    def addTestTeams():
        def reply(teams):
            print "Teams are : ", teams
        def reply3(team):
            print "New team is ", team
        def error(err):
            print "unable to add team : ", err

        print "Team types are : ", t.ListTeamTypes()

        t.ListTeams(reply, None)
        t.AddTeam("/org/freedesktop/Telepathy/Account/gabble/jabber/kakaroto_2etest1_40test_2ecollabora_2eco_2euk0", "xmpp_muc",
                  "teamgeist@conference.jabber.com", reply3, error)
        t.AddTeam("/org/freedesktop/Telepathy/Account/gabble/jabber/kakaroto_2etest2_40test_2ecollabora_2eco_2euk0", "xmpp_muc",
                  "teamgeist@conference.jabber.com", reply3, error)

        t.ListTeams(reply, None)


    def purgeTeams():
        def error(err):
            print "unable to purge team : ", err
        def reply(teams):
            def reply2(team):
                print "Purged team is ", team
            print "Teams are : ", teams
            for team in teams:
                print "***Purging ", team
                t.PurgeTeam(team, (reply2, team), error)
                print "***done ", teams
        t.ListTeams(reply, None)

    def testTopics():
        def reply(topics):
            print "Available topics : ", topics
        def error(err):
            print "Error ", err
        def reply2(teams):
            def reply3(topic):
                print "New topic : ", topic
            def reply4(topics):
                def reply5():
                    print "Deleted topic"

                for topic in topics:
                    print "Deleting topic ", topic
                    topic.team.DeleteTopic(topic, reply5, error)
            def reply6(team):
                print "Purged team is ", team
            team = None
            for team in teams:
                team.AddTopic('test_error', [], reply3, error)
                team.AddTopic('test1', [('tag', 0, 'teamgeist2')], reply3, error)
                team.AddTopic('test2', [('application', 0, 'opera'),
                               ('origin', 0, 'http://youtube.com')],
                              reply3, error)
            if team:
                team.ListTopics(reply, None)
                t.ListTopics(reply, None)
                #t.PurgeTeam(team, (reply6, team), error)
                t.ListTopics(reply, None)
                #t.ListTopics(reply4, None)

            #t.ListTeams(reply2, None)

        t.ListTopics(reply, None)
        t.ListTeams(reply2, None)

    def testEvents():
        import time
        def m():
            import struct
            t.MatchEvent(RemoteEvent(None, None, None, 1,
                 int(time.time() * 1000), 1, 2,
                 3, struct.pack("i", 7), []))
            t.MatchEvent(RemoteEvent(None, None, None, 5,
                 int(time.time() * 1000), 5, 6,
                 7, "blablabla", []))
            def reply(teams):
                for team in teams:
                    print team.GetEvents({":self:": [1,5]})
            t.ListTeams(reply, None)
            return False

        gobject.timeout_add(5000, m)


    import gobject
    import dbus
    import dbus.glib
    import logging
    #logging.basicConfig(level=logging.DEBUG)

    mainloop = gobject.MainLoop(is_running=True)
    dmainloop = dbus.mainloop.glib.DBusGMainLoop()
    dbus.set_default_main_loop(dmainloop)


    from dbus_objects import TeamgeistService
    ts = TeamgeistService()
    t = ts.teamgeist
    #t = Teamgeist()

    #testAccounts()
    #purgeTeams()
    #addTestTeams()
    #testTeams()
    #testTopics()
    #testEvents()
    #


    while mainloop.is_running():
        print 'running'
        try:
            mainloop.run()
        except KeyboardInterrupt:
            gobject.idle_add(quit, 1)
        print 'done\n\n'
