# -.- coding: utf-8 -.-

# Teamgeist
#
# Copyright (C) 2009 Collabora Limited
# *  Contact: Youness Alaoui
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# * Contributors:
# *   Youness Alaoui, Collabora Ltd.



import dbus.glib
import gobject
import logging

from tp_classes import *
from utils import do_reply
from signals import SignalDispatcher

logger = logging.getLogger('Teamgeist:TelepathyClient')

class ChannelDispatcher(BaseChannelDispatcher):
    pass

class ChannelRequest(BaseChannelRequest, SignalDispatcher):
    def __init__(self, bus_name = None, object_path = None, bus = None):
        BaseChannelRequest.__init__(self, bus_name, object_path, bus)
        SignalDispatcher.__init__(self)

    def Succeeded (self):
        print "ChannelRequest succeeded"
        self.emit_signal("Succeeded")

    def Failed(self, error, message):
        print "ChannelRequest Failed %s (%s)" % (error, message)
        self.emit_signal("Failed", error, message)

class TeamgeistClient(Client, ClientHandler, DBusProperties, SignalDispatcher):
    CLIENT_INTERFACE = 'org.freedesktop.Telepathy.Client'
    CLIENT_HANDLER_INTERFACE = 'org.freedesktop.Telepathy.Client.Handler'
    SERVICE_NAME = 'org.freedesktop.Telepathy.Client.TeamgeistTube'

    def __init__(self):
        """
        Initialise the connection manager.
        """
        SignalDispatcher.__init__(self)

        self.service_name = self.SERVICE_NAME
        object_path = "/" + self.service_name.replace(".", "/")
        Client.__init__(self,
                        dbus.service.BusName(self.service_name, dbus.Bus(),
                                             do_not_queue=True),
                        object_path)
        self._interfaces = set()
        self._interfaces.add(self.CLIENT_INTERFACE)
        self._interfaces.add(self.CLIENT_HANDLER_INTERFACE)
        DBusProperties.__init__(self)

        self._implement_property_get(self.CLIENT_INTERFACE, \
                {'Interfaces' : self.get_interfaces})
        self._implement_property_get(self.CLIENT_HANDLER_INTERFACE, \
                {'HandlerChannelFilter' : self.get_handler_channel_filter})
        self._implement_property_get(self.CLIENT_HANDLER_INTERFACE, \
                {'BypassApproval' : self.get_bypass_approval})
        self._implement_property_get(self.CLIENT_HANDLER_INTERFACE, \
                {'Capabilities' : self.get_capabilities})

        self.tubes = None
        self.own_tubes = None

        self._dispatcher = ChannelDispatcher()


    def _find_text_channel(self, connection, muc_id):
        channels = connection.channels
        for channel in channels.keys():
            p = channels[channel]
            if p[telepathy.CHANNEL + ".ChannelType"] == \
                    telepathy.CHANNEL_TYPE_TEXT and \
                    p[telepathy.CHANNEL + ".TargetHandleType"] == \
                    telepathy.HANDLE_TYPE_ROOM and \
                    p[telepathy.CHANNEL + ".TargetID"] == muc_id:
                return connection.get_related_object(Channel, channel)
        return None

    def request_muc(self, account, muc_id, reply_handler, error_handler):
        def account_ready():
            connection = account.connection
            if connection is None:
                do_reply(error_handler, Exception("Account is not connected"))
                return

            def connection_ready():
                channel = self._find_text_channel(connection, muc_id)
                if channel:
                    do_reply(reply_handler, channel)
                    return
                def ensured(req):
                    def error(emitter, error, message):
                        # TODO: if empathy is closed but you're connected
                        # there will be no handler so we'll get an error
                        # "Channel aborted"
                        do_reply(error_handler, Exception(message))
                    def succeeded(emitter):
                        channel = self._find_text_channel(connection, muc_id)
                        if channel:
                            do_reply(reply_handler, channel)
                        else:
                            do_reply(error_handler, Exception(
                                    "Could not find channel"))

                    cr = self._dispatcher.get_related_object(
                        ChannelRequest, req)
                    cr.connect("Succeeded", succeeded)
                    cr.connect("Failed", error)
                    cr.Proceed(ignore_reply=True)
                print "Ensuring Text channel on %s" % account
                self._dispatcher.EnsureChannel(account,
                                               {
                        telepathy.CHANNEL + '.ChannelType':
                            telepathy.CHANNEL_TYPE_TEXT,
                        telepathy.CHANNEL + '.TargetHandleType':
                            telepathy.HANDLE_TYPE_ROOM,
                        telepathy.CHANNEL + '.TargetID':
                            muc_id,
                        },
                                               0,
                                               "",
                                               reply_handler=ensured,
                                               error_handler=error_handler)
            connection.call_when_ready(connection_ready)
        account.call_when_ready(account_ready)


    def request_tube(self, account, muc_id, reply_handler, error_handler):
        def ensured(req):
            def error(emitter, error, message):
                do_reply(error_handler, Exception(message))
            def succeeded(emitter):
                do_reply(reply_handler)

            cr = self._dispatcher.get_related_object(
                ChannelRequest, req)
            cr.connect("Succeeded", succeeded)
            cr.connect("Failed", error)
            cr.Proceed(ignore_reply=True)
        print "Ensuring Tube channel on %s" % account
        self._dispatcher.CreateChannel(account,
                                       {
                telepathy.CHANNEL + '.ChannelType':
                    telepathy.CHANNEL_TYPE_DBUS_TUBE,
                telepathy.CHANNEL + '.TargetHandleType':
                    telepathy.HANDLE_TYPE_ROOM,
                telepathy.CHANNEL + '.TargetID':
                    muc_id,
                telepathy.CHANNEL_TYPE_DBUS_TUBE + '.ServiceName':
                    self.SERVICE_NAME,
                },
                                       0,
                                       self.service_name,
                                       reply_handler=ensured,
                                       error_handler=error_handler)


    # Properties
    def get_interfaces(self):
        return dbus.Array(sorted(self._interfaces), signature="s")

    def get_handler_channel_filter(self):
        return dbus.Array([{telepathy.CHANNEL + '.ChannelType':
                                telepathy.CHANNEL_TYPE_DBUS_TUBE,
                            telepathy.CHANNEL + '.TargetHandleType':
                                telepathy.HANDLE_TYPE_ROOM,
                            telepathy.CHANNEL_TYPE_DBUS_TUBE + '.ServiceName':
                                self.SERVICE_NAME,},], signature="a{sv}")

    def get_bypass_approval(self):
        return True

    def get_capabilities(self):
        return dbus.Array([], signature="s")

    # Methods
    @dbus.service.method('org.freedesktop.Telepathy.Client.Handler',
                         in_signature='ooa(oa{sv})aota{sv}', out_signature='')
    def HandleChannels(self, account, connection, channels,
                       requested, user_action_time, handler_info):
        for path, props in channels:
            print "HANDLE CHANNELS (%s) : %s" % (
                account, props[telepathy.CHANNEL + ".ChannelType"])

            if props[telepathy.CHANNEL + ".ChannelType"] == \
                    telepathy.CHANNEL_TYPE_DBUS_TUBE:
                service_name = connection.lstrip('/').replace('/', '.')
                muc_id = props[telepathy.CHANNEL + ".TargetID"]
                channel = DBusChannel(service_name, path)

                print "Got tube channel for muc %s on account %s" % (
                    muc_id, account)
                self.emit_signal("HandleTube", account, muc_id, channel, props)

            else:
                muc_id = props[telepathy.CHANNEL + ".TargetID"]
                print "Got %s channel for muc %s on account %s" % (
                    props[telepathy.CHANNEL + ".ChannelType"],
                    muc_id, account)

def run_mainloop():
    mainloop = gobject.MainLoop(is_running=True)
    dmainloop = dbus.mainloop.glib.DBusGMainLoop()
    dbus.set_default_main_loop(dmainloop)
    while mainloop.is_running():
        print 'running'
        try:
            mainloop.run()
        except KeyboardInterrupt:
            gobject.idle_add(quit, 1)
        print 'done\n\n'

if __name__ == "__main__":
    def _get_cl_cb(contacts):
        print "Contacts are : "
        for contact in contacts:
            print contact
        print ""

    def request_channel(conn, client):
        for account in conns._am.accounts.values():
            def account_ready(account):
                connection = account._connection
                if connection is not None:
                    client.open_tube(account, "test@conference.jabber.org")
            account.call_when_ready((account_ready, account))

    logging.basicConfig(level=logging.DEBUG)
    tm = TeamgeistClient()
    conns = TelepathyConnections()
    #conns._am.call_when_ready((request_channel, conns, tm))
    conns.GetContactLists((_get_cl_cb,), (_get_cl_cb,))
    run_mainloop()


# TODO : use gobject signals to notify of actions happening
# TODO : some type of singleton for my objects? so I can request objects from
# their paths
